#!/usr/bin/env python3
# MIT License
#
# ResourcesV2: fast, typed, lazy index over FIM AdvertizedTopology
#
# Author: Komal Thareja (kthare10@renci.org)

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Mapping, Optional, Tuple

from fim.user.topology import AdvertizedTopology

from fabrictestbed.util.site_v2 import (
    SiteV2, HostInfo, FacilityPortInfo, LinkInfo, LinkEndpoint,
    load_site, SwitchInfo
)


@dataclass
class ResourcesV2:
    """Normalized, high-performance index over a FIM AdvertizedTopology."""
    topology: AdvertizedTopology

    _sites: Dict[str, SiteV2] = field(default_factory=dict, init=False, repr=False)
    _hosts_by_site: Dict[str, Dict[str, HostInfo]] = field(default_factory=dict, init=False, repr=False)
    _switches_by_site: Dict[str, Dict[str, SwitchInfo]] = field(default_factory=dict, init=False, repr=False)

    # Lazy caches
    _facility_ports_by_site: Optional[Dict[str, Dict[str, FacilityPortInfo]]] = field(default=None, init=False, repr=False)
    _node_index: Optional[Dict[str, Tuple[str, str]]] = field(default=None, init=False, repr=False)
    _links_by_site_idx: Optional[Dict[str, List[int]]] = field(default=None, init=False, repr=False)
    _all_links: Optional[List[LinkInfo]] = field(default=None, init=False, repr=False)

    # -----------------------------------------------------
    def __post_init__(self) -> None:
        self._index_sites()

    # -----------------------------------------------------
    @property
    def sites(self) -> Mapping[str, SiteV2]:
        return self._sites

    def get_site(self, name: str) -> Optional[SiteV2]:
        s = self._sites.get(name)
        if not s:
            return None
        s._hosts_index = self._hosts_by_site.get(name, {})
        if self._facility_ports_by_site:
            s._facility_ports_index = self._facility_ports_by_site.get(name, {})
        if self._links_by_site_idx and self._all_links:
            s._links_index = {name: [self._all_links[i] for i in self._links_by_site_idx.get(name, [])]}
        return s

    def list_site_names(self) -> List[str]:
        return list(self._sites.keys())

    def list_hosts(self) -> List[HostInfo]:
        out: List[HostInfo] = []
        for site_hosts in self._hosts_by_site.values():
            out.extend(site_hosts.values())
        return out

    def list_facility_ports(self) -> List[FacilityPortInfo]:
        self._ensure_facility_ports()
        out: List[FacilityPortInfo] = []
        for fps in self._facility_ports_by_site.values():
            out.extend(fps.values())
        return out

    def list_links(self) -> List[LinkInfo]:
        self._ensure_links()
        return list(self._all_links or [])

    # -----------------------------------------------------
    # Indexing helpers
    # -----------------------------------------------------
    def _index_sites(self) -> None:
        for site_name, site in self.topology.sites.items():
            loaded_site, hosts, switches = load_site(site)
            if not hosts:
                continue

            # Store in master dicts
            self._sites[site_name] = loaded_site
            self._hosts_by_site[site_name] = hosts
            self._switches_by_site[site_name] = switches

            # HYDRATE the SiteV2 so it’s self-contained
            loaded_site._hosts_index = hosts
            loaded_site._switches_index = switches
            # facility ports / links remain lazy; they’ll be attached later on demand

    def _ensure_facility_ports(self) -> None:
        if self._facility_ports_by_site is not None:
            return
        fps_by_site: Dict[str, Dict[str, FacilityPortInfo]] = {}
        for fp in self.topology.facilities.values():
            site_name = fp.site
            site_bucket = fps_by_site.setdefault(site_name, {})
            for idx, iface in enumerate(fp.interface_list):
                labs = getattr(iface, "labels", None)
                local_name = getattr(labs, "local_name", None) if labs else None
                device_name = getattr(labs, "device_name", None) if labs else None

                # Try peers to fill in local/device name if missing (fablib behavior)
                if not local_name:
                    try:
                        for peer in iface.get_peers():
                            peer_labels = getattr(peer, "labels", None)
                            if peer_labels and getattr(peer_labels, "local_name", None):
                                local_name = getattr(peer_labels, "local_name", None)
                                if not device_name:
                                    device_name = getattr(peer_labels, "device_name", None)
                                break
                    except Exception:
                        pass

                label_allocations = None
                try:
                    label_allocations = iface.get_property("label_allocations")
                except Exception:
                    # property may not exist; best-effort only
                    label_allocations = None

                vlan_range = None
                if labs:
                    vr = getattr(labs, "vlan_range", None)
                    if vr:
                        vlan_range = str(vr)
                    else:
                        v = getattr(labs, "vlan", None)
                        if v is not None:
                            vlan_range = str([v])

                allocated_vlans = None
                if label_allocations:
                    alloc_vlan = getattr(label_allocations, "vlan", None)
                    if alloc_vlan is not None:
                        allocated_vlans = str(alloc_vlan)

                port_name = local_name or device_name or getattr(iface, "name", None)
                # Use interface-specific key so multiple interfaces per facility are kept
                # Ensure uniqueness even if node_id is reused across interfaces
                fp_key_parts = [
                    fp.name,
                    getattr(iface, "node_id", "") or "",
                    getattr(iface, "name", "") or "",
                    str(idx),
                ]
                fp_key = "|".join(fp_key_parts)

                site_bucket[fp_key] = FacilityPortInfo(
                    site=site_name,
                    name=fp.name,
                    port=port_name,
                    switch=getattr(iface, "node_id", None),
                    labels=labs,
                    vlans=vlan_range,
                    allocated_vlans=allocated_vlans,
                )
        self._facility_ports_by_site = fps_by_site

    def _build_node_index(self) -> None:
        if self._node_index is not None:
            return
        idx: Dict[str, Tuple[str, str]] = {}
        for site_name, site in self.topology.sites.items():
            for node_name, child in getattr(site, "children", {}).items():
                node_id = getattr(child, "node_id", None)
                if node_id:
                    idx[node_id] = (site_name, node_name)
        self._node_index = idx

    def _ensure_links(self) -> None:
        if self._all_links is not None:
            return
        self._build_node_index()
        node_idx = self._node_index

        all_links: List[LinkInfo] = []
        site_to_indices: Dict[str, List[int]] = {}

        for _, L in self.topology.links.items():
            site_names: Optional[Tuple[str, ...]] = None
            if getattr(L, "interface_list", None):
                first_iface = L.interface_list[0]
                iface_type = getattr(getattr(first_iface, "type", None), "name", None)
                iface_name = getattr(first_iface, "name", None)
                if iface_type == "TrunkPort" and iface_name:
                    parts = iface_name.split("_")
                    if parts and "HundredGig" not in parts[0]:
                        site_names = tuple(parts)

            endpoints: List[LinkEndpoint] = []
            for idx, iface in enumerate(L.interface_list):
                site_name, node_name = (None, None)
                if node_idx and getattr(iface, "node_id", None) in node_idx:
                    site_name, node_name = node_idx[iface.node_id]
                if site_name is None and site_names and idx < len(site_names):
                    site_name = site_names[idx]

                labs = getattr(iface, "labels", None)
                port = getattr(labs, "local_name", None) or getattr(labs, "device_name", None)
                port = port or getattr(iface, "name", None)

                endpoints.append(LinkEndpoint(
                    site=site_name,
                    node=node_name or getattr(iface, "node_id", None),
                    port=port
                ))

            li = LinkInfo(
                name=getattr(L, "node_id", None),
                layer=str(getattr(L, "layer", None)) if getattr(L, "layer", None) else None,
                labels=getattr(L, "labels", None),
                bandwidth=getattr(getattr(L, "capacities", None), "bw", None),
                allocated_bandwidth=getattr(getattr(L, "capacity_allocations", None), "bw", None),
                sites=site_names,
                endpoints=endpoints,
            )
            idx = len(all_links)
            all_links.append(li)
            for ep in endpoints:
                if ep.site:
                    site_to_indices.setdefault(ep.site, []).append(idx)

        self._all_links = all_links
        self._links_by_site_idx = site_to_indices



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Author Komal Thareja (kthare10@renci.org)
import datetime
import json
import logging
from typing import List, Dict, Any, Optional

import requests


class CoreApiError(Exception):
    """
    Core Exception
    """
    pass


class CoreApi:
    """
    Class implements functionality to interface with Core API
    """
    def __init__(self, core_api_host: str, token: str):
        if "https" not in core_api_host:
            self.api_server = f"https://{core_api_host}"
        else:
            self.api_server = core_api_host

        # Set the headers
        self.headers = {
            'Authorization': f'Bearer {token}',
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }

    def get_user_id(self) -> str:
        """
        Return User's uuid by querying via /whoami Core API
        
        @return User's uuid
        """
        url = f'{self.api_server}/whoami'
        response = requests.get(url, headers=self.headers)
        if response.status_code != 200:
            raise CoreApiError(f"Core API error occurred status_code: {response.status_code} "
                               f"message: {response.content}")

        logging.debug(f"GET WHOAMI Response : {response.json()}")

        return response.json().get("results")[0].get("uuid")

    def get_user_info_by_email(self, *, email: str) -> dict:
        if email is None:
            raise CoreApiError("Core API error email must be specified!")

        url = f'{self.api_server}/people?search={email}&exact_match=true&offset=0&limit=5'
        response = requests.get(url, headers=self.headers)
        if response.status_code != 200:
            raise CoreApiError(f"Core API error occurred status_code: {response.status_code} "
                               f"message: {response.content}")

        logging.debug(f"GET PEOPLE Response : {response.json()}")
        results = response.json().get("results")
        if len(results):
            return response.json().get("results")[0]

    def get_user_info(self, *, uuid: str = None, email: str = None) -> dict:
        """
        Return User's uuid by querying via /whoami Core API
        @param uuid User's uuid
        @param email email
        
        @return User's uuid and email
        """
        if email is not None:
            return self.get_user_info_by_email(email=email)

        if uuid is None and email is None:
            uuid = self.get_user_id()

        url = f'{self.api_server}/people/{uuid}?as_self=true'
        response = requests.get(url, headers=self.headers)
        if response.status_code != 200:
            raise CoreApiError(f"Core API error occurred status_code: {response.status_code} "
                               f"message: {response.content}")

        logging.debug(f"GET WHOAMI Response : {response.json()}")

        return response.json().get("results")[0]

    def __get_user_project_by_id(self, *, project_id: str) -> list:
        """
        Return user's project identified by project id
        
        @param project_id project id
        @return list of the projects
        """
        url = f"{self.api_server}/projects/{project_id}"
        response = requests.get(url, headers=self.headers)

        if response.status_code != 200:
            raise CoreApiError(f"Core API error occurred status_code: {response.status_code} "
                               f"message: {response.content}")

        logging.debug(f"GET Project Response : {response.json()}")

        return response.json().get("results")

    def __get_user_projects(self, *, project_name: str = None, uuid: str = None) -> list:
        """
        Return user's project identified by project name
        
        @param project_name project name
        @return list of the projects
        """
        offset = 0
        limit = 50
        if uuid is None:
            uuid = self.get_user_id()
        result = []
        total_fetched = 0

        while True:
            if project_name is not None:
                url = f"{self.api_server}/projects?search={project_name}&offset={offset}&limit={limit}" \
                      f"&person_uuid={uuid}&sort_by=name&order_by=asc"
            else:
                url = f"{self.api_server}/projects?offset={offset}&limit={limit}&person_uuid={uuid}" \
                      f"&sort_by=name&order_by=asc"

            response = requests.get(url, headers=self.headers)

            if response.status_code != 200:
                raise CoreApiError(f"Core API error occurred status_code: {response.status_code} "
                                   f"message: {response.content}")

            logging.debug(f"GET Project Response : {response.json()}")

            size = response.json().get("size")
            total = response.json().get("total")
            projects = response.json().get("results")

            total_fetched += size

            for x in projects:
                result.append(x)

            if total_fetched == total:
                break
            offset = size
            limit += limit

        return result

    def get_user_projects(self, project_name: str = "all", project_id: str = "all", uuid: str = None) -> List[dict]:
        """
        Get User's projects either identified by project name, project id or all
        @param project_name Project name
        @param project_id Project Id
        @param uuid User Id
        @return list of projects
        """
        ret_val = []
        if project_id is not None and project_id != "all":
            projects = self.__get_user_project_by_id(project_id=project_id)
        elif project_name is not None and project_name != "all":
            projects = self.__get_user_projects(project_name=project_name, uuid=uuid)
        else:
            projects = self.__get_user_projects()

        for p in projects:
            expires_on = p.get("expires_on")
            if expires_on is not None:
                expires_on_dt = datetime.datetime.fromisoformat(expires_on)
                now = datetime.datetime.now(tz=datetime.timezone.utc)
                if now > expires_on_dt:
                    # Do not include the expired project in the token for "all" get slices
                    if project_id is not None and project_id.lower() == "all":
                        continue
                    # Do not include the expired project in the token for "all" get slices
                    elif project_name is not None and project_name.lower() == "all":
                        continue
                    # Fail a request of the token for an expired token
                    else:
                        raise CoreApiError(f"Project {p.get('name')} is expired!")

            project_memberships = p.get("memberships")

            if not project_memberships["is_member"] and not project_memberships["is_creator"] and \
                    not project_memberships["is_owner"]:
                raise CoreApiError(f"User is not a member of Project: {p.get('uuid')}")
            project = {
                "name": p.get("name"),
                "uuid": p.get("uuid")
            }

            # Only pass tags and membership when token is requested for a specific project
            if project_id.lower() != "all":
                project["tags"] = p.get("tags")
                project["memberships"] = p.get("memberships")

            ret_val.append(project)

        if len(ret_val) == 0:
            raise CoreApiError(f"User is not a member of Project: {project_id}:{project_name}")

        return ret_val

    def get_ssh_keys(self, *, uuid: str = None, email: str = None) -> list:
        """
        Return SSH Keys, given user's uuid
        @param uuid: uuid
        @param email:

        @return list of ssh keys
        """
        if email is not None:
            user_info = self.get_user_info(email=email)
            if user_info is None:
                raise CoreApiError(f"Core API error {email} not found")
            uuid = user_info.get("uuid")
        elif uuid is None:
            uuid = self.get_user_id()

        url = f'{self.api_server}/sshkeys?person_uuid={uuid}'
        response = requests.get(url, headers=self.headers)
        if response.status_code != 200:
            raise CoreApiError(f"Core API error occurred status_code: {response.status_code} "
                               f"message: {response.content}")

        logging.debug(f"GET SSH Keys Response : {response.json()}")
        return response.json().get("results")

    def create_ssh_keys(self, key_type: str, description: str,
                        comment: str = "ssh-key-via-api", store_pubkey: bool = True) -> list:
        """
        Create SSH Keys for a user
        @param description: Key Description
        @param comment: Comment
        @param store_pubkey: Flag indicating if public key should be saved
        @param key_type: Key Type (sliver or bastion)

        @return list of ssh keys
        """
        data = {
            "comment": comment,
            "description": description,
            "keytype": key_type,
            "store_pubkey": store_pubkey
        }

        # Make a POST request to the core-api API
        response = requests.post(f'{self.api_server}/sshkeys', headers=self.headers, data=json.dumps(data))

        if response.status_code != 200:
            raise CoreApiError(f"Core API error occurred status_code: {response.status_code} "
                               f"message: {response.content}")

        logging.debug(f"POST SSH Keys Response : {response.json()}")
        return response.json().get("results")

    def list_quotas(self, project_uuid: str, offset: int = 0, limit: int = 200) -> List[dict]:
        """
        Fetch all quotas from the API.

        Parameters:
        project_uuid (str): The UUID of the project for which quotas are fetched.
        offset (int, optional): The starting point for fetching records. Defaults to 0.
        limit (int, optional): The maximum number of records to fetch. Defaults to 200.

        Returns:
        List[dict]: A list of quota records.
        """
        params = {"project_uuid": project_uuid,
                  "offset": offset,
                  "limit": limit}

        response = requests.get(f'{self.api_server}/quotas', headers=self.headers, params=params)
        self.raise_for_status(response=response)
        logging.debug(f"GET Quotas Response : {response.json()}")
        return response.json().get("results")

    def create_quota(self, project_uuid: str, resource_type: str, resource_unit: str,
                     quota_used: int = 0, quota_limit: int = 0) -> List[dict]:
        """
        Send a POST request to create a new quota.

        Parameters:
        project_uuid (str): The ID of the project to which the quota belongs.
        resource_type (str): The type of resource (e.g., CPU, RAM).
        resource_unit (str): The unit of the resource (e.g., cores, GB).
        quota_used (int, optional): The amount of resource currently used. Defaults to 0.
        quota_limit (int, optional): The limit for the resource. Defaults to 0.

        Returns:
        List[dict]: The created quota details.
        """
        data = {
            "project_uuid": project_uuid,
            "resource_type": resource_type,
            "resource_unit": resource_unit,
            "quota_used": quota_used,
            "quota_limit": quota_limit
        }
        response = requests.post(f'{self.api_server}/quotas', headers=self.headers, json=data)
        self.raise_for_status(response=response)
        logging.debug(f"POST Quotas Response : {response.json()}")
        return response.json().get("results")

    def get_quota(self, uuid: str) -> List[dict]:
        """
        Send a GET request to retrieve a quota by UUID.

        Parameters:
        uuid (str): The UUID of the quota to retrieve.

        Returns:
        List[dict]: The retrieved quota details.
        """
        response = requests.get(f'{self.api_server}/quotas/{uuid}', headers=self.headers)
        self.raise_for_status(response=response)
        logging.debug(f"GET Quotas Response : {response.json()}")
        return response.json().get("results")

    def update_quota_usage(self, uuid: str, project_uuid: str, quota_used: float):
        """
        Send a PUT request to update a quota usage by UUID.

        Parameters:
        uuid (str): The UUID of the quota to update.
        project_uuid (str): The ID of the project to which the quota belongs.
        quota_used (int, optional): The amount of resource currently used.

        Returns:
        List[dict]: The updated quota details.
        """
        current_quota = self.get_quota(uuid=uuid)[0]

        value = current_quota.get("quota_used") + quota_used
        if value < 0:
            value = 0

        data = {
            "project_uuid": project_uuid,
            "resource_type": current_quota.get("resource_type"),
            "resource_unit": current_quota.get("resource_unit"),
            "quota_used": value,
            "quota_limit": current_quota.get("quota_limit")
        }
        response = requests.put(f'{self.api_server}/quotas/{uuid}', headers=self.headers, json=data)
        self.raise_for_status(response=response)
        logging.debug(f"PUT Quotas Response : {response.json()}")
        return response.json().get("results")

    def update_quota(self, uuid: str, project_uuid: str, resource_type: str, resource_unit: str,
                     quota_used: float = 0, quota_limit: float = 0):
        """
        Send a PUT request to update a quota by UUID.

        Parameters:
        uuid (str): The UUID of the quota to update.
        project_uuid (str): The ID of the project to which the quota belongs.
        resource_type (str): The type of resource (e.g., CPU, RAM).
        resource_unit (str): The unit of the resource (e.g., cores, GB).
        quota_used (int, optional): The amount of resource currently used. Defaults to 0.
        quota_limit (int, optional): The limit for the resource. Defaults to 0.

        Returns:
        List[dict]: The updated quota details.
        """
        data = {
            "project_uuid": project_uuid,
            "resource_type": resource_type,
            "resource_unit": resource_unit,
            "quota_used": quota_used,
            "quota_limit": quota_limit
        }
        response = requests.put(f'{self.api_server}/quotas/{uuid}', headers=self.headers, json=data)
        self.raise_for_status(response=response)
        logging.debug(f"PUT Quotas Response : {response.json()}")
        return response.json().get("results")

    def delete_quota(self, uuid: str):
        """
        Send a DELETE request to delete a quota by UUID.

        Parameters:
        uuid (str): The UUID of the quota to delete.

        Returns:
        """
        response = requests.delete(f'{self.api_server}/quotas/{uuid}', headers=self.headers)
        self.raise_for_status(response=response)
        logging.debug(f"DEL Quotas Response : {response.json()}")

    def list_storage(self, offset: int = 0, limit: int = 200) -> List[dict]:
        """
        Fetch all storage volumes from the API.

        Parameters:
        offset (int, optional): The starting point for fetching records. Defaults to 0.
        limit (int, optional): The maximum number of records to fetch. Defaults to 200.

        Returns:
        List[dict]: A list of storage volume records.
        """
        params = {"offset": offset, "limit": limit}
        response = requests.get(f'{self.api_server}/storage', headers=self.headers, params=params)
        self.raise_for_status(response=response)
        logging.debug(f"GET Storage Response : {response.json()}")
        return response.json().get("results")

    def get_storage(self, uuid: str) -> List[dict]:
        """
        Send a GET request to retrieve a storage volume by UUID.

        Parameters:
        uuid (str): The UUID of the storage volume to retrieve.

        Returns:
        List[dict]: The retrieved storage volume details.
        """
        response = requests.get(f'{self.api_server}/storage/{uuid}', headers=self.headers)
        self.raise_for_status(response=response)
        logging.debug(f"GET Storage Response : {response.json()}")
        return response.json().get("results")

    def get_ssh_key(self, uuid: str) -> List[dict]:
        """
        Send a GET request to retrieve an SSH key by UUID.

        Parameters:
        uuid (str): The UUID of the SSH key to retrieve.

        Returns:
        List[dict]: The retrieved SSH key details.
        """
        response = requests.get(f'{self.api_server}/sshkeys/{uuid}', headers=self.headers)
        self.raise_for_status(response=response)
        logging.debug(f"GET SSH Key Response : {response.json()}")
        return response.json().get("results")

    def get_bastion_keys(self, secret: str, since_date: str) -> List[dict]:
        """
        Send a GET request to retrieve bastion keys.

        Parameters:
        secret (str): The secret required for authentication.
        since_date (str): The date from which to fetch bastion keys.

        Returns:
        List[dict]: The retrieved bastion keys.
        """
        params = {"secret": secret, "since_date": since_date}
        response = requests.get(f'{self.api_server}/bastionkeys', headers=self.headers, params=params)
        self.raise_for_status(response=response)
        logging.debug(f"GET Bastion Keys Response : {response.json()}")
        return response.json().get("results")

    def get_project(self, uuid: str) -> List[dict]:
        """
        Send a GET request to retrieve a project by UUID.

        Parameters:
        uuid (str): The UUID of the project to retrieve.

        Returns:
        List[dict]: The retrieved project details.
        """
        response = requests.get(f'{self.api_server}/projects/{uuid}', headers=self.headers)
        self.raise_for_status(response=response)
        logging.debug(f"GET Project Response : {response.json()}")
        return response.json().get("results")

    def list_project_users(self, project_uuid: str) -> List[dict]:
        """
        List users for a project.

        :param project_uuid: Project UUID
        :return: List of user records with roles
        """
        if not project_uuid:
            raise CoreApiError("project_uuid is required")

        project_details = self.get_project(uuid=project_uuid)
        project_obj: Optional[dict] = None
        if isinstance(project_details, list) and project_details:
            project_obj = project_details[0]
        elif isinstance(project_details, dict):
            project_obj = project_details

        def _add_user(bucket: Dict[str, Dict[str, Any]], person: Dict[str, Any], role_label: str) -> None:
            if not isinstance(person, dict):
                return
            user_uuid = person.get("uuid") or person.get("user_uuid") or person.get("person_uuid")
            if not user_uuid:
                return
            rec = bucket.setdefault(
                user_uuid,
                {
                    "project_uuid": project_uuid,
                    "user_uuid": user_uuid,
                    "email": person.get("email") or person.get("username"),
                    "name": person.get("name") or person.get("full_name") or person.get("fullname"),
                    "roles": [],
                },
            )
            if role_label and role_label not in rec["roles"]:
                rec["roles"].append(role_label)

        users_index: Dict[str, Dict[str, Any]] = {}
        if isinstance(project_obj, dict):
            for person in project_obj.get("project_members", []) or []:
                _add_user(users_index, person, "member")
            for person in project_obj.get("project_owners", []) or []:
                _add_user(users_index, person, "owner")
            for person in project_obj.get("project_creators", []) or []:
                _add_user(users_index, person, "creator")
            for person in project_obj.get("token_holders", []) or []:
                _add_user(users_index, person, "token_holder")
            lead = project_obj.get("project_lead")
            if lead:
                _add_user(users_index, lead, "lead")
            # Fallback keys
            for person in project_obj.get("members") or []:
                _add_user(users_index, person, "member")
            for person in project_obj.get("users") or []:
                _add_user(users_index, person, "member")

        users_out: List[Dict[str, Any]] = []
        for rec in users_index.values():
            rec["role"] = ",".join(rec["roles"]) if rec.get("roles") else None
            users_out.append(rec)

        return users_out

    @staticmethod
    def raise_for_status(response: requests.Response):
        """
        Checks the response status and raises an ArtifactManagerError if the request was unsuccessful.

        :param response: The response object returned from the API request.
        :raises ArtifactManagerError: If the response contains an HTTP error.
        """
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            try:
                message = response.json()
            except json.JSONDecodeError:
                message = {"message": "Unknown error occurred while processing the request."}

            raise CoreApiError(f"Error {response.status_code}: {e}. Message: {message}")


if __name__ == '__main__':
    project_id = ""
    token = ""
    core_api = CoreApi(core_api_host="alpha-6.fabric-testbed.net", token=token)

    quotas = core_api.list_quotas(project_uuid=project_id)
    print(f"Fetching quotas: {json.dumps(quotas,indent=4)}")
'''
    resources = ["core", "ram", "disk"]
    if len(quotas) == 0:
        for r in resources:
            core_api.create_quota(project_uuid=project_id, resource_type=r, resource_unit="hours",
                                  quota_limit=100, quota_used=0)
            print(f"Created quota for {r}")

    for q in quotas:
        core_api.update_quota(uuid=q.get("uuid"), project_uuid=q.get("project_uuid"),
                              quota_limit=q.get("quota_limit"), quota_used=q.get("quota_used") + 1,
                              resource_type=q.get("resource_type"),
                              resource_unit=q.get("resource_unit"))
        qq = core_api.get_quota(uuid=q.get("uuid"))
        print(f"Updated quota: {qq}")

    for q in quotas:
        print(f"Deleting quota: {q.get('uuid')}")
        core_api.delete_quota(uuid=q.get("uuid"))

    quotas = core_api.list_quotas(project_uuid="74a5b28b-c1a2-4fad-882b-03362dddfa71")
    print(f"Quotas after deletion!: {quotas}")


    #core_api.create_quota(project_uuid="74a5b28b-c1a2-4fad-882b-03362dddfa71",
    #                      resource_type="disk", resource_unit="hours", quota_limit=100)
'''



#!/usr/bin/env python3
# MIT License
#
# Author: Komal Thareja (kthare10@renci.org)

from __future__ import annotations

import json
import logging
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

# Optional: if present, we can do GUI login to get a cookie.
try:
    from fabric_cm.credmgr.session_helper import SessionHelper  # type: ignore
    _HAS_SESSION_HELPER = True
except Exception:
    _HAS_SESSION_HELPER = False


# =========================
# DTOs
# =========================

@dataclass(slots=True)
class TokenDTO:
    token_hash: str
    created_at: str
    expires_at: str
    state: str
    created_from: Optional[str] = None
    comment: Optional[str] = None
    id_token: Optional[str] = None
    refresh_token: Optional[str] = None

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "TokenDTO":
        return TokenDTO(
            token_hash=d.get("token_hash"),
            created_at=d.get("created_at"),
            expires_at=d.get("expires_at"),
            state=d.get("state"),
            created_from=d.get("created_from"),
            comment=d.get("comment"),
            id_token=d.get("id_token"),
            refresh_token=d.get("refresh_token"),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "token_hash": self.token_hash,
            "created_at": self.created_at,
            "expires_at": self.expires_at,
            "state": self.state,
            "created_from": self.created_from,
            "comment": self.comment,
            "id_token": self.id_token,
            "refresh_token": self.refresh_token,
        }


@dataclass(slots=True)
class DecodedTokenDTO:
    token: Dict[str, Any]

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DecodedTokenDTO":
        # API returns { ..., "token": {...} }
        token = d.get("token") or {}
        return DecodedTokenDTO(token=token)

    def to_dict(self) -> Dict[str, Any]:
        return {"token": self.token}


# =========================
# Errors
# =========================

class CredMgrError(Exception):
    pass


class CredMgrValidationError(CredMgrError):
    pass


class CredMgrHTTPError(CredMgrError):
    def __init__(self, status_code: int, message: str, body_preview: str = ""):
        super().__init__(f"{status_code}: {message}\n{body_preview}")
        self.status_code = status_code
        self.body_preview = body_preview


# =========================
# Core client (merged)
# =========================

class CredmgrClient:
    """
    Single-class client that merges the HTTP layer and the higher-level token helpers.

    Features:
    - requests + retries
    - structured logging
    - optional GUI login cookie via SessionHelper (if importable)
    - DTO/dataclass conversions
    - return_fmt = "dict" | "dto" for MCP friendliness
    - on-disk token.json helpers
    """

    TIME_FORMAT = "%Y-%m-%d %H:%M:%S %z"

    def __init__(
        self,
        credmgr_host: str,
        *,
        timeout: float = 30.0,
        retries: int = 3,
        backoff_factor: float = 0.3,
        logger: Optional[logging.Logger] = None,
        http_debug: bool = False,
        cookie_provider: Optional[Callable[[], str]] = None,  # alternative to SessionHelper
        cookie_name: str = "fabric-service",
    ):
        if not credmgr_host:
            raise CredMgrError("credmgr_host must be provided")

        if not credmgr_host.startswith("http"):
            base = f"https://{credmgr_host}"
        else:
            base = credmgr_host
        if not base.endswith("/"):
            base += "/"

        # OpenAPI shows base like .../credmgr/
        if not base.endswith("credmgr/"):
            base += "credmgr/"

        self.base_url = base
        self.timeout = timeout
        self.session = self._build_session(retries, backoff_factor)
        self.log = logger or logging.getLogger("fabric.credmgr")
        self.http_debug = http_debug
        self.cookie_provider = cookie_provider
        self.cookie_name = cookie_name
        self._cookie: Optional[str] = None  # cache between calls

    # -------------
    # Session + HTTP
    # -------------

    @staticmethod
    def _build_session(retries: int, backoff: float) -> requests.Session:
        s = requests.Session()
        retry = Retry(
            total=retries,
            connect=retries,
            read=retries,
            status=retries,
            backoff_factor=backoff,
            status_forcelist=(429, 500, 502, 503, 504),
            allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE"]),
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry, pool_connections=10, pool_maxsize=20)
        s.mount("https://", adapter)
        s.mount("http://", adapter)
        return s

    def _req(
        self,
        method: str,
        path: str,
        *,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Union[Dict[str, Any], List[Tuple[str, Any]]]] = None,
        json_body: Optional[Dict[str, Any]] = None,
        data: Optional[Union[str, bytes]] = None,
        cookie: Optional[str] = None,
    ) -> requests.Response:
        url = self.base_url + path.lstrip("/")
        hdrs = {"Accept": "application/json"}
        if headers:
            hdrs.update(headers)

        cookies = {}
        if cookie:
            cookies[self.cookie_name] = cookie

        t0 = time.perf_counter()
        resp = self.session.request(
            method=method,
            url=url,
            headers=hdrs,
            params=params,
            json=json_body,
            data=data,
            cookies=cookies if cookies else None,
            timeout=self.timeout,
        )
        dt = (time.perf_counter() - t0) * 1000.0
        self.log.info(
            "credmgr_http_call",
            extra={
                "event": "http_call",
                "http_method": method,
                "url_path": path,
                "status_code": resp.status_code,
                "elapsed_ms": round(dt, 2),
                "resp_len": len(resp.content or b""),
            },
        )

        if not (200 <= resp.status_code < 300):
            if self.http_debug:
                self.log.debug(
                    "credmgr_http_failure_dump",
                    extra={
                        "event": "http_failure",
                        "http_method": method,
                        "url": url,
                        "status_code": resp.status_code,
                        "request_headers": hdrs,
                        "request_params": params,
                        "request_json": json_body,
                        "response_text": resp.text[:2000],
                    },
                )
            raise CredMgrHTTPError(resp.status_code, "HTTP request failed", resp.text[:2000])
        return resp

    @staticmethod
    def _json(resp: requests.Response) -> Dict[str, Any]:
        try:
            return resp.json() if resp.content else {}
        except Exception as e:
            raise CredMgrError(f"Failed to parse JSON: {e}; status={resp.status_code}, text={resp.text[:500]}")

    # -------------
    # Cookie helpers
    # -------------

    def _resolve_cookie(
        self,
        *,
        cookie: Optional[str],
        use_gui_login: bool,
        browser_name: str,
        base_ui_url: Optional[str] = None,
        wait_timeout: int = 500,
        wait_interval: int = 5,
    ) -> Optional[str]:
        """
        Resolve a cookie in priority:
        1) explicit cookie argument
        2) cached self._cookie
        3) cookie_provider() if set
        4) GUI login via SessionHelper (if available and use_gui_login=True)

        base_ui_url defaults to "<scheme>://<host>/" inferred from self.base_url when not provided.
        """
        if cookie:
            self._cookie = cookie
            return cookie
        if self._cookie:
            return self._cookie
        if self.cookie_provider:
            try:
                ck = self.cookie_provider()
                if ck:
                    self._cookie = ck
                    return ck
            except Exception as e:
                self.log.warning("cookie_provider_failed", extra={"error": str(e)})

        if use_gui_login:
            if not _HAS_SESSION_HELPER:
                raise CredMgrError("SessionHelper is not available; cannot perform GUI login.")
            # infer https://<host>/
            if not base_ui_url:
                # self.base_url like https://host/credmgr/ -> want https://host/
                parts = self.base_url.split("/credmgr/")[0]
                if not parts.endswith("/"):
                    parts += "/"
                base_ui_url = parts
            session = SessionHelper(
                url=base_ui_url,
                cookie_name=self.cookie_name,
                wait_timeout=wait_timeout,
                wait_interval=wait_interval,
            )
            ck = session.login(browser_name=browser_name)
            self._cookie = ck
            return ck

        return None

    # =========================
    # CredMgr API
    # =========================

    def version(self) -> Dict[str, Any]:
        resp = self._req("GET", "/version")
        return self._json(resp)

    def certs(self) -> Dict[str, Any]:
        resp = self._req("GET", "/certs")
        return self._json(resp)

    def create(
        self,
        *,
        scope: str = "all",
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        lifetime_hours: int = 4,
        comment: str = "Create Token via GUI",
        # cookie acquisition options:
        cookie: Optional[str] = None,
        use_gui_login: bool = False,
        browser_name: str = "chrome",
        base_ui_url: Optional[str] = None,
        wait_timeout: int = 500,
        wait_interval: int = 5,
        # output options:
        file_path: Optional[Union[str, Path]] = None,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[TokenDTO] | List[Dict[str, Any]]:
        """
        Create token(s) using a GUI-authenticated cookie.
        Provide either `cookie` or set `use_gui_login=True` (requires SessionHelper).
        """
        if not project_id and not project_name:
            raise CredMgrValidationError("project_id or project_name must be specified")

        ck = self._resolve_cookie(
            cookie=cookie,
            use_gui_login=use_gui_login,
            browser_name=browser_name,
            base_ui_url=base_ui_url,
            wait_timeout=wait_timeout,
            wait_interval=wait_interval,
        )
        if not ck:
            raise CredMgrValidationError("No cookie available to call /tokens/create")

        params: Dict[str, Any] = {
            "scope": scope,
            "lifetime": lifetime_hours,
            "comment": comment,
        }
        if project_id:
            params["project_id"] = project_id
        if project_name:
            params["project_name"] = project_name

        resp = self._req("POST", "/tokens/create", params=params, cookie=ck)
        payload = self._json(resp)
        data = payload.get("data") or []

        if file_path:
            # save the first token (most useful) to disk preserving the familiar layout
            first = data[0] if data else {}
            self.save_file(file_path, first)

        tokens = [TokenDTO.from_dict(x) for x in data]
        return [t.to_dict() for t in tokens] if return_fmt == "dict" else tokens

    def refresh(
        self,
        *,
        refresh_token: str,
        scope: str = "all",
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        file_path: Optional[Union[str, Path]] = None,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[TokenDTO] | List[Dict[str, Any]]:
        if not refresh_token:
            raise CredMgrValidationError("refresh_token must be provided")
        if not project_id and not project_name:
            raise CredMgrValidationError("project_id or project_name must be specified")

        params: Dict[str, Any] = {"scope": scope}
        if project_id:
            params["project_id"] = project_id
        if project_name:
            params["project_name"] = project_name

        body = {"refresh_token": refresh_token}
        resp = self._req("POST", "/tokens/refresh", params=params, json_body=body)
        payload = self._json(resp)
        data = payload.get("data") or []

        if file_path and data:
            self.save_file(file_path, data[0])

        tokens = [TokenDTO.from_dict(x) for x in data]
        return [t.to_dict() for t in tokens] if return_fmt == "dict" else tokens

    def revoke(
        self,
        *,
        id_token: str,
        token_type: Literal["refresh", "identity"],
        refresh_token: Optional[str] = None,
        token_hash: Optional[str] = None,
    ) -> None:
        """
        Revoke a refresh token or an identity token (by hash).
        """
        if token_type == "refresh" and not refresh_token:
            raise CredMgrValidationError("refresh_token is required when token_type='refresh'")
        if token_type == "identity" and not token_hash:
            raise CredMgrValidationError("token_hash is required when token_type='identity'")

        headers = {"Authorization": f"Bearer {id_token}", "Content-Type": "application/json"}
        post_body = {"type": token_type, "token": refresh_token if token_type == "refresh" else token_hash}
        resp = self._req("POST", "/tokens/revokes", headers=headers, json_body=post_body)
        _ = self._json(resp)  # no-content schema in spec

    def delete(self, *, id_token: str, token_hash: str) -> None:
        headers = {"Authorization": f"Bearer {id_token}"}
        self._req("DELETE", f"/tokens/delete/{token_hash}", headers=headers)

    def delete_all(self, *, id_token: str) -> None:
        headers = {"Authorization": f"Bearer {id_token}"}
        self._req("DELETE", "/tokens/delete", headers=headers)

    def tokens(
        self,
        *,
        id_token: str,
        project_id: Optional[str] = None,
        token_hash: Optional[str] = None,
        expires: Optional[str] = None,  # TIME_FORMAT on server side
        states: Optional[List[str]] = None,
        limit: int = 200,
        offset: int = 0,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[TokenDTO] | List[Dict[str, Any]]:
        headers = {"Authorization": f"Bearer {id_token}"}
        params: List[Tuple[str, Any]] = [("limit", limit), ("offset", offset)]
        if token_hash:
            params.append(("token_hash", token_hash))
        if project_id:
            params.append(("project_id", project_id))
        if expires:
            params.append(("expires", expires))
        if states:
            for s in states:
                params.append(("states", s))

        resp = self._req("GET", "/tokens", headers=headers, params=params)
        payload = self._json(resp)
        data = payload.get("data") or []
        tokens = [TokenDTO.from_dict(x) for x in data]
        return [t.to_dict() for t in tokens] if return_fmt == "dict" else tokens

    def revoke_list(self, *, id_token: str, project_id: Optional[str] = None) -> List[str]:
        headers = {"Authorization": f"Bearer {id_token}"}
        params: Dict[str, Any] = {}
        if project_id:
            params["project_id"] = project_id
        resp = self._req("GET", "/tokens/revoke_list", headers=headers, params=params)
        payload = self._json(resp)
        return payload.get("data") or []

    def validate(
        self,
        *,
        id_token: str,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> DecodedTokenDTO | Dict[str, Any]:
        headers = {"Content-Type": "application/json"}
        body = {"type": "identity", "token": id_token}
        resp = self._req("POST", "/tokens/validate", headers=headers, json_body=body)
        payload = self._json(resp)
        dto = DecodedTokenDTO.from_dict(payload)
        return dto.to_dict() if return_fmt == "dict" else dto

    # =========================
    # File helpers
    # =========================

    @staticmethod
    def save_file(file_path: Union[str, Path], token_data: Dict[str, Any]) -> None:
        """
        Save token data to a JSON file.

        :param file_path: Path to the token file.
        :param token_data: Token dictionary to save.
        """
        path = Path(file_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as f:
            json.dump(token_data, f, indent=2)

    @staticmethod
    def load_file(file_path: Union[str, Path]) -> Dict[str, Any]:
        """
        Load token data from a JSON file.

        :param file_path: Path to the token file.
        :return: Token dictionary.
        """
        path = Path(file_path)
        if not path.exists():
            return {}
        with open(path, "r") as f:
            return json.load(f)




#!/usr/bin/env python3
# MIT License
#
# Author: Komal Thareja (kthare10@renci.org)

from __future__ import annotations

import json
import logging
import os
import time
from datetime import datetime, timezone, timedelta
from pathlib import Path

from typing import Any, Dict, List, Literal, Optional, Union

from fim.user.topology import AdvertizedTopology, Topology

from fabrictestbed.external_api.credmgr_client import CredmgrClient
from fabrictestbed.external_api.orchestrator_client import OrchestratorClient, SliverDTO, SliceDTO, PoaDataDTO
from fabrictestbed.external_api.core_api import CoreApi
from fabrictestbed.fabric_manager import FabricManagerException
from fabrictestbed.topology_query_api import TopologyQueryAPI
from fabrictestbed.util.constants import Constants
from fabrictestbed.util.utils import Utils


class FabricManagerV2Exception(Exception):
    """Custom exception for FabricManagerV2-related errors."""
    pass


# =========================
# FabricManagerV2 façade
# =========================

class FabricManagerV2(TopologyQueryAPI):
    """
    One-stop façade over CredmgrClient + OrchestratorClient with integrated token management.

    Supports two token modes:
    1. **File-based (legacy)** — provide ``token_location`` path to a JSON file.
       Tokens are read/written on disk and auto-refreshed when near expiry.
    2. **In-memory (MCP-friendly)** — provide ``id_token`` and optionally
       ``refresh_token``; file I/O is skipped unless ``no_write=False``.

    All orchestrator methods can either:
    - Accept explicit ``id_token`` parameter, or
    - Use ``ensure_valid_id_token()`` when ``id_token`` is not passed

    :param credmgr_host: Credential Manager hostname (e.g., ``cm.fabric-testbed.net``).
    :param orchestrator_host: Orchestrator hostname (e.g., ``orchestrator.fabric-testbed.net``).
    :param core_api_host: Optional Core API hostname for user/project lookups.
    :param token_location: Path to tokens JSON file (file-based mode). Use ``None`` for in-memory mode.
    :param id_token: In-memory FABRIC ID token (JWT) for MCP mode.
    :param refresh_token: In-memory refresh token for MCP mode (optional).
    :param project_id: Optional project UUID for scoping token refresh.
    :param project_name: Optional project name for scoping token refresh.
    :param scope: Token scope for refresh flows (default: ``"all"``).
    :param no_write: If ``True``, never write tokens to disk (recommended for MCP).
    :param auto_refresh: If ``True``, automatically refresh tokens when near expiry (default: ``True``).
    :param logger: Optional custom logger.
    :param http_debug: Enable HTTP debug logging.
    """

    def __init__(
        self,
        *,
        credmgr_host: str,
        orchestrator_host: str,
        core_api_host: Optional[str] = None,
        token_location: Optional[str] = None,
        id_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        scope: str = "all",
        no_write: bool = False,
        auto_refresh: bool = True,
        logger: Optional[logging.Logger] = None,
        http_debug: bool = False,
    ):
        super().__init__(orchestrator_host=orchestrator_host, logger=logger, http_debug=http_debug)
        self.credmgr = CredmgrClient(credmgr_host, http_debug=http_debug, logger=self.log)
        self.orch = OrchestratorClient(orchestrator_host, http_debug=http_debug, logger=self.log)
        self.core_api_host = core_api_host

        # Token management attributes
        self.token_location = token_location
        self.no_write = no_write
        self.auto_refresh = auto_refresh
        self.scope = scope
        self.project_id = project_id
        self.project_name = project_name
        self.user_id: Optional[str] = None
        self.user_email: Optional[str] = None

        # In-memory token storage
        self._tokens: Dict[str, Any] = {}
        if id_token:
            self._tokens["id_token"] = id_token
        if refresh_token:
            self._tokens["refresh_token"] = refresh_token

        # Load tokens from file if not provided in-memory
        if not self._tokens.get("id_token"):
            self._load_tokens()

        # Extract project/user info from token
        try:
            self._extract_info_from_token()
        except Exception as e:
            self.log.debug(f"Non-fatal: could not decode id_token for project/user info: {e}")

    # ------------------------------------------------------------------
    # Token Management
    # ------------------------------------------------------------------

    def _load_tokens(self) -> None:
        """Load tokens from file or environment."""
        if self._tokens.get("id_token"):
            return

        if self.token_location and os.path.exists(self.token_location):
            try:
                self._tokens = CredmgrClient.load_file(self.token_location)
            except Exception as e:
                self.log.warning(f"Failed to read token file {self.token_location}: {e}")
                self._tokens = {}

        # Try environment variable as fallback
        refresh_token = self._tokens.get("refresh_token") or os.environ.get(Constants.CILOGON_REFRESH_TOKEN)

        # Auto-refresh on load if we have refresh token
        if self.auto_refresh and refresh_token and not self._tokens.get("id_token"):
            try:
                self.refresh_tokens(refresh_token=refresh_token)
            except Exception as e:
                self.log.warning(f"Auto-refresh on load failed: {e}")

    def _extract_info_from_token(self) -> None:
        """Decode id_token to populate project/user attributes if missing."""
        if self.project_id and self.project_name and self.user_id and self.user_email:
            return

        tok = self.get_id_token()
        if not tok:
            return

        try:
            decoded = self.credmgr.validate(id_token=tok, return_fmt="dict")
            token_info = decoded.get("token", {})
            self.project_id = self.project_id or token_info.get("project_id")
            self.project_name = self.project_name or token_info.get("project_name")
            self.user_id = self.user_id or token_info.get("uuid")
            self.user_email = self.user_email or token_info.get("email")
        except Exception as e:
            self.log.debug(f"Failed to decode token: {e}")

    def get_id_token(self) -> Optional[str]:
        """
        Retrieve the current ID token.

        :return: The ID token string if available, otherwise ``None``.
        """
        return self._tokens.get("id_token")

    def get_refresh_token(self) -> Optional[str]:
        """
        Retrieve the current refresh token.

        :return: The refresh token string if available, otherwise ``None``.
        """
        return self._tokens.get("refresh_token")

    def id_token_expires_at(self) -> Optional[datetime]:
        """
        Return when the ID token expires, if determinable.

        :return: Datetime of expiration or ``None`` if unavailable.
        """
        # Check stored expires_at first
        expires_at = self._tokens.get("expires_at")
        if expires_at:
            try:
                return datetime.strptime(expires_at, CredmgrClient.TIME_FORMAT)
            except Exception:
                pass

        # Decode from token claims
        tok = self.get_id_token()
        if tok:
            try:
                decoded = self.credmgr.validate(id_token=tok, return_fmt="dict")
                token_info = decoded.get("token", {})
                exp = token_info.get("exp")
                if exp:
                    return datetime.fromtimestamp(int(exp), timezone.utc)
            except Exception:
                pass
        return None

    def refresh_tokens(self, *, refresh_token: Optional[str] = None) -> tuple:
        """
        Refresh the tokens via Credential Manager.

        If ``no_write=False`` and ``token_location`` is set, saves refreshed tokens to file.

        :param refresh_token: Refresh token to use. If not provided, uses stored refresh token.
        :return: Tuple of (new_id_token, new_refresh_token)
        :raises FabricManagerV2Exception: If the refresh operation fails.
        """
        rtok = refresh_token or self.get_refresh_token()
        if not rtok:
            raise FabricManagerV2Exception("No refresh token available")

        if not self.project_id and not self.project_name:
            raise FabricManagerV2Exception("project_id or project_name must be set for token refresh")

        try:
            # Determine file path for saving
            file_path = None
            if not self.no_write and self.token_location:
                file_path = self.token_location

            tokens = self.credmgr.refresh(
                refresh_token=rtok,
                scope=self.scope,
                project_id=self.project_id,
                project_name=self.project_name,
                file_path=file_path,
                return_fmt="dict",
            )

            if tokens and len(tokens) > 0:
                first_token = tokens[0]
                self._tokens["id_token"] = first_token.get("id_token")
                self._tokens["refresh_token"] = first_token.get("refresh_token")
                self._tokens["expires_at"] = first_token.get("expires_at")
                self._tokens["created_at"] = first_token.get("created_at")

                return (
                    self._tokens.get("id_token"),
                    self._tokens.get("refresh_token"),
                )

            raise FabricManagerV2Exception("Refresh returned empty token list")

        except Exception as e:
            raise FabricManagerV2Exception(f"Token refresh failed: {e}")

    def ensure_valid_id_token(self, id_token: Optional[str] = None) -> str:
        """
        Ensure a non-expired ID token, refreshing if possible.

        If ``id_token`` is provided explicitly, returns it directly.
        Otherwise, checks the stored token and refreshes if near expiry (within 30 minutes).

        :param id_token: Optional explicit token to use (bypasses refresh logic).
        :return: A valid ID token.
        :raises FabricManagerV2Exception: If no token is available or refresh fails.
        """
        # If explicit token provided, use it directly
        if id_token:
            return id_token

        tok = self.get_id_token()
        if not tok:
            raise FabricManagerV2Exception(
                "No id_token available. Provide id_token parameter, set token_location, or pass tokens during init."
            )

        if self.auto_refresh:
            expires_at = self.id_token_expires_at()
            if expires_at and datetime.now(timezone.utc) >= (expires_at - timedelta(minutes=30)):
                rtok = self.get_refresh_token()
                if rtok:
                    self.log.info("Token near expiry, auto-refreshing...")
                    self.refresh_tokens(refresh_token=rtok)
                    return self.get_id_token()
                else:
                    self.log.warning("Token near expiry but no refresh token available")
        else:
            expires_at = self.id_token_expires_at()
            if expires_at and datetime.now(timezone.utc) >= expires_at:
                raise FabricManagerV2Exception("Token expired and auto_refresh is disabled")

        return tok

    def get_user_id(self) -> Optional[str]:
        """Retrieve the user ID from the token."""
        if not self.user_id and self.get_id_token():
            self._extract_info_from_token()
        return self.user_id

    def get_user_email(self) -> Optional[str]:
        """Retrieve the user email from the token."""
        if not self.user_email and self.get_id_token():
            self._extract_info_from_token()
        return self.user_email

    def get_project_name(self) -> Optional[str]:
        """Retrieve the project name from the token."""
        if not self.project_name and self.get_id_token():
            self._extract_info_from_token()
        return self.project_name

    def get_project_id(self) -> Optional[str]:
        """Retrieve the project id from the token."""
        return self.project_id

    def get_user_info(self, *, uuid: str = None, email: str = None) -> dict:
        """
        Retrieve user info by email or uuid.

        :param uuid: User's uuid.
        :type uuid: str
        :param email: User's email address.
        :type email: str
        :return: User info dict (shape per Core API).
        :rtype: dict
        :raises FabricManagerException: On Core API errors.
        """
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.get_user_info(uuid=uuid, email=email)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    # -------- Token helpers (proxy) --------

    def tokens_create(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Create new FABRIC tokens.

        Proxies to CredmgrClient.create with dict return format for MCP compatibility.

        Args:
            **kwargs: Keyword arguments passed to CredmgrClient.create.

        Returns:
            List of token dictionaries with token details.
        """
        # Same signature as CredmgrV3.create; return dicts by default (MCP friendly)
        return self.credmgr.create(return_fmt="dict", **kwargs)  # type: ignore[arg-type]

    def tokens_refresh(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Refresh existing FABRIC tokens.

        Args:
            **kwargs: Keyword arguments passed to CredmgrClient.refresh.

        Returns:
            List of refreshed token dictionaries.
        """
        return self.credmgr.refresh(return_fmt="dict", **kwargs)  # type: ignore[arg-type]

    def tokens_validate(self, *, id_token: str) -> Dict[str, Any]:
        """
        Validate a FABRIC ID token.

        Args:
            id_token: The FABRIC ID token to validate.

        Returns:
            Dictionary containing validation result and token details.
        """
        return self.credmgr.validate(id_token=id_token, return_fmt="dict")  # type: ignore[return-value]

    def tokens_list(self, **kwargs) -> List[Dict[str, Any]]:
        """
        List all FABRIC tokens for the user.

        Args:
            **kwargs: Keyword arguments passed to CredmgrClient.tokens.

        Returns:
            List of token dictionaries.
        """
        return self.credmgr.tokens(return_fmt="dict", **kwargs)  # type: ignore[arg-type]

    def tokens_revoke(self, **kwargs) -> None:
        """
        Revoke a FABRIC token.

        Args:
            **kwargs: Keyword arguments passed to CredmgrClient.revoke.
        """
        return self.credmgr.revoke(**kwargs)

    def tokens_delete(self, **kwargs) -> None:
        """
        Delete a FABRIC token.

        Args:
            **kwargs: Keyword arguments passed to CredmgrClient.delete.
        """
        return self.credmgr.delete(**kwargs)

    def tokens_delete_all(self, **kwargs) -> None:
        """
        Delete all FABRIC tokens for the user.

        Args:
            **kwargs: Keyword arguments passed to CredmgrClient.delete_all.
        """
        return self.credmgr.delete_all(**kwargs)

    def tokens_revoke_list(self, **kwargs) -> List[str]:
        """
        Get list of revoked tokens.

        Args:
            **kwargs: Keyword arguments passed to CredmgrClient.revoke_list.

        Returns:
            List of revoked token identifiers.
        """
        return self.credmgr.revoke_list(**kwargs)

    # -------- Orchestrator helpers --------
    def create_slice(
        self,
        *,
        name: str,
        graph_model: str,
        ssh_keys: List[str],
        id_token: Optional[str] = None,
        lifetime: Optional[int] = None,
        lease_start_time: Optional[str] = None,
        lease_end_time: Optional[str] = None,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[Union[Dict[str, Any], SliverDTO]]:
        """
        Create a new FABRIC slice.

        Args:
            name: Name of the slice to create.
            graph_model: Slice topology graph model (GRAPHML, JSON, etc.).
            ssh_keys: List of SSH public keys for slice access.
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            lifetime: Optional slice lifetime in days.
            lease_start_time: Optional lease start time (UTC format).
            lease_end_time: Optional lease end time (UTC format).
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            List of sliver dictionaries or DTO objects representing the created slice resources.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.create(
            token=token,
            slice_name=name,
            slice_graph=graph_model,
            ssh_keys=ssh_keys,
            lifetime=lifetime,
            lease_start_time=lease_start_time,
            lease_end_time=lease_end_time,
            return_fmt=return_fmt,
        )

    def modify_slice(
        self, *, slice_id: str, graph_model: str, id_token: Optional[str] = None, return_fmt: Literal["dict", "dto"] = "dict"
    ) -> List[Union[Dict[str, Any], SliverDTO]]:
        """
        Modify an existing FABRIC slice topology.

        Args:
            slice_id: UUID of the slice to modify.
            graph_model: Updated slice topology graph model.
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            List of sliver dictionaries or DTO objects with modification results.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.modify(token=token, slice_id=slice_id, slice_graph=graph_model, return_fmt=return_fmt)

    def accept_modify(
        self, *, slice_id: str, id_token: Optional[str] = None, return_fmt: Literal["dict", "dto"] = "dict"
    ) -> Union[Dict[str, Any], SliceDTO]:
        """
        Accept pending slice modifications.

        Args:
            slice_id: UUID of the slice with pending modifications.
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            Slice dictionary or DTO object with updated state.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.modify_accept(token=token, slice_id=slice_id, return_fmt=return_fmt)

    def renew_slice(self, *, slice_id: str, lease_end_time: str, id_token: Optional[str] = None) -> None:
        """
        Renew a FABRIC slice lease.

        Args:
            slice_id: UUID of the slice to renew.
            lease_end_time: New lease end time (UTC format).
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.renew(token=token, slice_id=slice_id, new_lease_end_time=lease_end_time)

    def delete_slice(self, *, slice_id: Optional[str] = None, id_token: Optional[str] = None) -> None:
        """
        Delete a FABRIC slice.

        Args:
            slice_id: Optional UUID of the slice to delete.
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.delete(token=token, slice_id=slice_id)

    def list_slices(
        self,
        *,
        id_token: Optional[str] = None,
        states: Optional[List[str]] = None,
        exclude_states: Optional[List[str]] = None,
        name: Optional[str] = None,
        search: Optional[str] = None,
        exact_match: bool = False,
        as_self: bool = True,
        limit: int = 200,
        offset: int = 0,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[Union[Dict[str, Any], SliceDTO]]:
        """
        List FABRIC slices with optional filtering.

        Args:
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            states: Optional list of slice states to include (e.g., ["StableError", "StableOK"]). Allowed values: (Nascent, Configuring, StableOK,
                     StableError, ModifyOK, ModifyError, Closing, Dead).
            exclude_states: Optional list of slice states to exclude (e.g., for fetching active slices set exclude_states=["Closing", "Dead"]).
            name: Optional slice name filter.
            search: Optional search string for slice names.
            exact_match: If True, match slice name exactly; if False, use substring match.
            as_self: If True, list only user's own slices; if False, list all accessible slices.
            limit: Maximum number of slices to return (default: 200).
            offset: Pagination offset (default: 0).
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            List of slice dictionaries or DTO objects.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.slices(
            token=token,
            includes=states,
            excludes=exclude_states,
            name=name,
            search=search,
            exact_match=exact_match,
            as_self=as_self,
            limit=limit,
            offset=offset,
            return_fmt=return_fmt,
        )

    def get_slice(
        self,
        *,
        slice_id: str,
        id_token: Optional[str] = None,
        graph_format: Literal["GRAPHML", "JSON_NODELINK", "CYTOSCAPE", "NONE"] = "GRAPHML",
        as_self: bool = True,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> Union[Dict[str, Any], SliceDTO]:
        """
        Get details of a specific FABRIC slice.

        Args:
            slice_id: UUID of the slice to retrieve.
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            graph_format: Format for the slice topology graph (GRAPHML, JSON_NODELINK, CYTOSCAPE, or NONE).
            as_self: If True, retrieve as owner; if False, retrieve with delegated access.
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            Slice dictionary or DTO object with full details.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.get_slice(token=token, slice_id=slice_id, graph_format=graph_format,
                                   as_self=as_self, return_fmt=return_fmt)

    def list_slivers(
        self, *, slice_id: str, id_token: Optional[str] = None, as_self: bool = True, return_fmt: Literal["dict", "dto"] = "dict"
    ) -> List[Union[Dict[str, Any], SliverDTO]]:
        """
        List all slivers (resource allocations) in a slice.

        Args:
            slice_id: UUID of the slice containing the slivers.
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            as_self: If True, list as owner; if False, list with delegated access.
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            List of sliver dictionaries or DTO objects.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.slivers(token=token, slice_id=slice_id, as_self=as_self, return_fmt=return_fmt)

    def get_sliver(
        self, *, slice_id: str, sliver_id: str, id_token: Optional[str] = None, as_self: bool = True, return_fmt: Literal["dict", "dto"] = "dict"
    ) -> List[Union[Dict[str, Any], SliverDTO]]:
        """
        Get details of a specific sliver.

        Args:
            slice_id: UUID of the slice containing the sliver.
            sliver_id: UUID of the sliver to retrieve.
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            as_self: If True, retrieve as owner; if False, retrieve with delegated access.
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            List containing the sliver dictionary or DTO object.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.slivers(token=token, slice_id=slice_id, sliver_id=sliver_id, as_self=as_self,
                                 return_fmt=return_fmt)

    def poa_create(
        self,
        *,
        sliver_id: str,
        operation: Literal["cpuinfo", "numainfo", "cpupin", "numatune", "reboot", "addkey", "removekey", "rescan"],
        id_token: Optional[str] = None,
        vcpu_cpu_map: Optional[List[Dict[str, str]]] = None,
        node_set: Optional[List[str]] = None,
        keys: Optional[List[Dict[str, str]]] = None,
        bdf: Optional[List[str]] = None,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[Union[Dict[str, Any], PoaDataDTO]]:
        """
        Create a Perform Operational Action (POA) on a sliver.

        POAs allow runtime operations on deployed resources like CPU pinning,
        NUMA tuning, reboots, SSH key management, and device rescanning.

        Args:
            sliver_id: UUID of the sliver to perform the action on.
            operation: The POA operation to perform (cpuinfo, numainfo, cpupin, numatune,
                      reboot, addkey, removekey, rescan).
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            vcpu_cpu_map: Optional list of vCPU-to-CPU mapping dictionaries for cpupin operation.
            node_set: Optional list of NUMA node identifiers for numatune operation.
            keys: Optional list of SSH key dictionaries for addkey/removekey operations.
            bdf: Optional list of Bus:Device.Function identifiers for rescan operation.
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            List of POA result dictionaries or DTO objects.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.poa(
            token=token,
            sliver_id=sliver_id,
            operation=operation,
            vcpu_cpu_map=vcpu_cpu_map,
            node_set=node_set,
            keys=keys,
            bdf=bdf,
            return_fmt=return_fmt,
        )

    def poa(
        self,
        *,
        sliver_id: str,
        operation: str,
        id_token: Optional[str] = None,
        vcpu_cpu_map: Optional[List[Dict[str, str]]] = None,
        node_set: Optional[List[str]] = None,
        keys: Optional[List[Dict[str, str]]] = None,
        bdf: Optional[List[str]] = None,
        return_fmt: Literal["dict", "dto"] = "dict",
        wait: bool = True,
        retry: int = 20,
        interval_seconds: int = 10,
    ) -> Union[List[Union[Dict[str, Any], PoaDataDTO]], str, Dict[str, Any]]:
        """
        Generic POA wrapper to invoke an operational action on a sliver.

        :param sliver_id: UUID of the sliver to act on.
        :param operation: POA operation (e.g., "reboot", "addkey", "removekey").
        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :param vcpu_cpu_map: Optional mapping for cpupin.
        :param node_set: Optional node set for numatune.
        :param keys: Optional list of key dicts for addkey/removekey (e.g., {"key": "<pub>", "comment": "..."}).
        :param bdf: Optional list of BDF identifiers for rescan.
        :param return_fmt: "dict" or "dto".
        :param wait: If True, poll until POA completes; else return submission response.
        :param retry: Max polling attempts (default 20).
        :param interval_seconds: Seconds between polls (default 10).
        :return: POA result list or final status/info if wait=True.
        """
        token = self.ensure_valid_id_token(id_token)
        submission = self.orch.poa(
            token=token,
            sliver_id=sliver_id,
            operation=operation,
            vcpu_cpu_map=vcpu_cpu_map,
            node_set=node_set,
            keys=keys,
            bdf=bdf,
            return_fmt=return_fmt,
        )
        # submission may be list of PoaDataDTO or dicts
        poa_id = None
        if isinstance(submission, list) and submission:
            first = submission[0]
            poa_id = getattr(first, "poa_id", None) or first.get("poa_id") if isinstance(first, dict) else None
        elif isinstance(submission, dict):
            poa_id = submission.get("poa_id")

        if not wait or not poa_id:
            return submission

        attempt = 0
        terminal_states = {"Success", "Failed"}
        latest = submission
        while attempt < retry:
            attempt += 1
            latest = self.orch.get_poas(
                token=token,
                poa_id=poa_id,
                limit=1,
                offset=0,
                return_fmt=return_fmt,
            )
            if isinstance(latest, list) and latest:
                entry = latest[0]
                state = getattr(entry, "state", None) or (entry.get("state") if isinstance(entry, dict) else None)
                if state in terminal_states:
                    if state == "Failed":
                        err = getattr(entry, "error", None) or (entry.get("error") if isinstance(entry, dict) else None)
                        raise ValueError(f"POA {poa_id}/{operation} failed: {err}")
                    info = getattr(entry, "info", None) or (entry.get("info") if isinstance(entry, dict) else None)
                    if isinstance(info, dict) and info.get(operation) is not None:
                        return info.get(operation)
                    return entry
            time.sleep(interval_seconds)

        return latest

    def poa_get(
        self,
        *,
        id_token: Optional[str] = None,
        sliver_id: Optional[str] = None,
        poa_id: Optional[str] = None,
        limit: int = 200,
        offset: int = 0,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[Union[Dict[str, Any], PoaDataDTO]]:
        """
        Retrieve POA (Perform Operational Action) status and results.

        Args:
            id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
            sliver_id: Optional UUID of the sliver to filter POAs by.
            poa_id: Optional UUID of a specific POA to retrieve.
            limit: Maximum number of POAs to return (default: 200).
            offset: Pagination offset (default: 0).
            return_fmt: Return format - "dict" for dictionaries or "dto" for DTO objects.

        Returns:
            List of POA dictionaries or DTO objects with status and results.
        """
        token = self.ensure_valid_id_token(id_token)
        return self.orch.get_poas(
            token=token, sliver_id=sliver_id, poa_id=poa_id, limit=limit, offset=offset, return_fmt=return_fmt
        )

    # -------- Storage helpers (Core API) --------
    def list_storage(
        self, *, id_token: Optional[str] = None, offset: int = 0, limit: int = 200
    ) -> List[Dict[str, Any]]:
        """
        List all storage volumes.

        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :param offset: Pagination offset (default: 0).
        :param limit: Maximum number of records to fetch (default: 200).
        :return: List of storage volume dictionaries.
        """
        if not self.core_api_host:
            raise ValueError("core_api_host must be provided during initialization to use storage methods")
        token = self.ensure_valid_id_token(id_token)
        core_api = CoreApi(core_api_host=self.core_api_host, token=token)
        return core_api.list_storage(offset=offset, limit=limit)

    def get_storage(self, *, uuid: str, id_token: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        Get a specific storage volume by UUID.

        :param uuid: Storage volume UUID.
        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :return: Storage volume details.
        """
        if not self.core_api_host:
            raise ValueError("core_api_host must be provided during initialization to use storage methods")
        token = self.ensure_valid_id_token(id_token)
        core_api = CoreApi(core_api_host=self.core_api_host, token=token)
        return core_api.get_storage(uuid=uuid)

    # -------- Project helpers (Core API) --------
    def get_project_info(
        self,
        *,
        id_token: Optional[str] = None,
        project_name: str = "all",
        project_id: str = "all",
        uuid: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Retrieve project info for the current user (or specified uuid) via Core API.

        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :param project_name: Project name filter (default "all").
        :param project_id: Project id filter (default "all").
        :param uuid: Optional user UUID; Core API infers current user if omitted.
        :return: List of matching project records.
        """
        if not project_name and not project_id:
            raise ValueError("project_name or project_id must be provided")
        if not self.core_api_host:
            raise ValueError("core_api_host must be provided during initialization to use project methods")
        token = self.ensure_valid_id_token(id_token)
        core_api = CoreApi(core_api_host=self.core_api_host, token=token)
        return core_api.get_user_projects(uuid=uuid, project_name=project_name, project_id=project_id)

    def list_project_users(
        self,
        *,
        project_uuid: str,
        id_token: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        List users in a project (via Core API).

        :param project_uuid: Project UUID to inspect.
        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :return: List of user records with roles.
        """
        if not project_uuid:
            raise ValueError("project_uuid is required")
        if not self.core_api_host:
            raise ValueError("core_api_host must be provided during initialization to use project methods")

        token = self.ensure_valid_id_token(id_token)
        core_api = CoreApi(core_api_host=self.core_api_host, token=token)
        return core_api.list_project_users(project_uuid=project_uuid)

    def get_user_keys(
        self,
        *,
        user_uuid: str,
        id_token: Optional[str] = None,
        key_type_filter: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Fetch SSH/public keys for a specific user (person_uuid).

        :param user_uuid: User UUID (person_uuid).
        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :param key_type_filter: Optional key type filter (e.g., "sliver", "bastion").
        :return: List of key records.
        """
        if not self.core_api_host:
            raise ValueError("core_api_host must be provided during initialization to use project methods")

        token = self.ensure_valid_id_token(id_token)
        core_api = CoreApi(core_api_host=self.core_api_host, token=token)
        keys = core_api.get_ssh_keys(uuid=user_uuid)

        if not key_type_filter:
            return keys

        def _kt(k: Dict[str, Any]) -> str:
            return str(
                k.get("keytype")
                or k.get("fabric_key_type")
                or k.get("fabric_keytype")
                or k.get("key_type")
                or ""
            ).lower()

        return [k for k in keys if _kt(k) == key_type_filter.lower()]

    def os_reboot(
        self,
        *,
        sliver_id: str,
        id_token: Optional[str] = None,
        return_fmt: Literal["dict", "dto"] = "dict",
    ) -> List[Union[Dict[str, Any], PoaDataDTO]]:
        """
        Issue a POA reboot on a sliver (NodeSliver only).

        :param sliver_id: Sliver UUID to reboot (NodeSliver only).
        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :param return_fmt: Return format - "dict" or "dto".
        :return: POA results.
        """
        return self.poa(
            id_token=id_token,
            sliver_id=sliver_id,
            operation="reboot",
            return_fmt=return_fmt,
        )

    def add_public_key(
        self,
        *,
        sliver_id: str,
        id_token: Optional[str] = None,
        sliver_key_name: Optional[str] = None,
        email: Optional[str] = None,
        sliver_public_key: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Add a public key to a sliver via POA addkey (NodeSliver only).

        :param sliver_id: Sliver UUID to act on (NodeSliver only).
        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :param sliver_key_name: Sliver key comment/name (portal) to fetch from Core API.
        :param email: Optional email for fetching another user's key by name.
        :param sliver_public_key: Raw public key string to add (if provided, overrides key lookup). MUST be "{ssh_key_type} {public_key}".
        :return: POA results from orchestrator.
        """
        if not sliver_key_name and not sliver_public_key:
            raise ValueError("Either sliver_key_name or sliver_public_key must be provided")

        token = self.ensure_valid_id_token(id_token)
        key_material = sliver_public_key
        if key_material and " " not in key_material.strip():
            raise ValueError(
                "sliver_public_key must include key type and key material, e.g., 'ecdsa-sha2-nistp256 AAAA...=='"
            )
        if not key_material and sliver_key_name:
            core_api = CoreApi(core_api_host=self.core_api_host, token=token)
            key_list = core_api.get_ssh_keys(email=email)
            for item in key_list or []:
                if sliver_key_name == item.get("comment"):
                    key_material = f"{item.get('ssh_key_type')} {item.get('public_key')}"
                    break
            if not key_material:
                raise ValueError(f"Sliver key '{sliver_key_name}' not found")

        if not key_material:
            raise ValueError("Key must be provided")

        key_comment = "addkey-by-api"
        if sliver_key_name or email:
            key_comment = f"{key_comment}:{sliver_key_name or ''}:{email or ''}"
        keys = [{"key": key_material, "comment": key_comment}]

        return self.poa(
            id_token=token,
            sliver_id=sliver_id,
            operation="addkey",
            keys=keys,
        )

    def remove_public_key(
        self,
        *,
        sliver_id: str,
        id_token: Optional[str] = None,
        sliver_key_name: Optional[str] = None,
        email: Optional[str] = None,
        sliver_public_key: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Remove a public key from a sliver via POA removekey (NodeSliver only).

        :param sliver_id: Sliver UUID to act on (NodeSliver only).
        :param id_token: Optional FABRIC ID token. If not provided, uses stored/auto-refreshed token.
        :param sliver_key_name: Sliver key comment/name (portal) to fetch from Core API.
        :param email: Optional email for fetching another user's key by name.
        :param sliver_public_key: Raw public key string to remove (if provided, overrides key lookup). MUST be "{ssh_key_type} {public_key}".
        :return: POA results from orchestrator.
        """
        if not sliver_key_name and not sliver_public_key:
            raise ValueError("Either sliver_key_name or sliver_public_key must be provided")

        token = self.ensure_valid_id_token(id_token)
        key_material = sliver_public_key
        if key_material and " " not in key_material.strip():
            raise ValueError(
                "sliver_public_key must include key type and key material, e.g., 'ecdsa-sha2-nistp256 AAAA...=='"
            )
        if not key_material and sliver_key_name:
            core_api = CoreApi(core_api_host=self.core_api_host, token=token)
            key_list = core_api.get_ssh_keys(email=email)
            for item in key_list or []:
                if sliver_key_name == item.get("comment"):
                    key_material = f"{item.get('ssh_key_type')} {item.get('public_key')}"
                    break
            if not key_material:
                raise ValueError(f"Sliver key '{sliver_key_name}' not found")

        if not key_material:
            raise ValueError("Key must be provided")

        key_comment = "removekey-by-api"
        if sliver_key_name or email:
            key_comment = f"{key_comment}:{sliver_key_name or ''}:{email or ''}"
        keys = [{"key": key_material, "comment": key_comment}]

        return self.poa(
            id_token=token,
            sliver_id=sliver_id,
            operation="removekey",
            keys=keys,
        )



if __name__ == "__main__":
    id_token = ""
    mgr = FabricManagerV2(credmgr_host="cm.fabric-testbed.net",
                          orchestrator_host="orchestrator.fabric-testbed.net",)

    start = time.perf_counter()
    sites = mgr.query_sites(id_token=id_token)
    print(json.dumps(sites, indent=2))
    print("Query sites time:", time.perf_counter() - start)


    start = time.perf_counter()
    hosts = mgr.query_hosts(id_token=id_token)
    print(json.dumps(hosts, indent=2))
    print("Query hosts time:", time.perf_counter() - start)

    start = time.perf_counter()
    fps = mgr.query_facility_ports(id_token=id_token)
    print(json.dumps(fps, indent=2))
    print("Query Facility Ports time:", time.perf_counter() - start)

    start = time.perf_counter()
    fps = mgr.query_links(id_token=id_token)
    print(json.dumps(fps, indent=2))
    print("Query Links time:", time.perf_counter() - start)



#!/usr/bin/env python3
# MIT License
#
# Author: Komal Thareja (kthare10@renci.org)

from __future__ import annotations

import json
import logging
import time
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple, Union, Literal

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from fim.user import GraphFormat
from fim.user.topology import ExperimentTopology, AdvertizedTopology


# ===========================
# Dataclasses (DTOs)
# ===========================

@dataclass(slots=True)
class SliceDTO:
    slice_id: str
    name: Optional[str] = None
    state: Optional[str] = None
    model: Optional[str] = None
    lease_start_time: Optional[str] = None
    lease_end_time: Optional[str] = None
    project_id: Optional[str] = None
    project_name: Optional[str] = None
    graph_id: Optional[str] = None
    owner_user_id: Optional[str] = None
    owner_email: Optional[str] = None

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SliceDTO":
        return SliceDTO(
            slice_id=d.get("slice_id"),
            name=d.get("name"),
            state=d.get("state"),
            model=d.get("model"),
            lease_start_time=d.get("lease_start_time"),
            lease_end_time=d.get("lease_end_time"),
            project_id=d.get("project_id"),
            project_name=d.get("project_name"),
            graph_id=d.get("graph_id"),
            owner_user_id=d.get("owner_user_id"),
            owner_email=d.get("owner_email"),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "slice_id": self.slice_id,
            "name": self.name,
            "state": self.state,
            "model": self.model,
            "lease_start_time": self.lease_start_time,
            "lease_end_time": self.lease_end_time,
            "project_id": self.project_id,
            "project_name": self.project_name,
            "graph_id": self.graph_id,
            "owner_user_id": self.owner_user_id,
            "owner_email": self.owner_email,
        }


@dataclass(slots=True)
class SliverDTO:
    sliver_id: str
    slice_id: str
    graph_node_id: Optional[str] = None
    sliver_type: Optional[str] = None
    sliver: Optional[Dict[str, Any]] = None
    lease_start_time: Optional[str] = None
    lease_end_time: Optional[str] = None
    state: Optional[str] = None
    pending_state: Optional[str] = None
    join_state: Optional[str] = None
    notice: Optional[str] = None
    owner_user_id: Optional[str] = None
    owner_email: Optional[str] = None

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "SliverDTO":
        return SliverDTO(
            sliver_id=d.get("sliver_id"),
            slice_id=d.get("slice_id"),
            graph_node_id=d.get("graph_node_id"),
            sliver_type=d.get("sliver_type"),
            sliver=d.get("sliver"),
            lease_start_time=d.get("lease_start_time"),
            lease_end_time=d.get("lease_end_time"),
            state=d.get("state"),
            pending_state=d.get("pending_state"),
            join_state=d.get("join_state"),
            notice=d.get("notice"),
            owner_user_id=d.get("owner_user_id"),
            owner_email=d.get("owner_email"),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "sliver_id": self.sliver_id,
            "slice_id": self.slice_id,
            "graph_node_id": self.graph_node_id,
            "sliver_type": self.sliver_type,
            "sliver": self.sliver,
            "lease_start_time": self.lease_start_time,
            "lease_end_time": self.lease_end_time,
            "state": self.state,
            "pending_state": self.pending_state,
            "join_state": self.join_state,
            "notice": self.notice,
            "owner_user_id": self.owner_user_id,
            "owner_email": self.owner_email,
        }


@dataclass(slots=True)
class PoaDataDTO:
    poa_id: Optional[str] = None
    operation: Optional[str] = None
    state: Optional[str] = None
    sliver_id: Optional[str] = None
    slice_id: Optional[str] = None
    error: Optional[str] = None
    info: Optional[Dict[str, Any]] = None

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "PoaDataDTO":
        return PoaDataDTO(
            poa_id=d.get("poa_id"),
            operation=d.get("operation"),
            state=d.get("state"),
            sliver_id=d.get("sliver_id"),
            slice_id=d.get("slice_id"),
            error=d.get("error"),
            info=d.get("info"),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "poa_id": self.poa_id,
            "operation": self.operation,
            "state": self.state,
            "sliver_id": self.sliver_id,
            "slice_id": self.slice_id,
            "error": self.error,
            "info": self.info,
        }


# ===========================
# Exceptions
# ===========================

class OrchestratorError(Exception):
    """Base error for orchestrator client."""


class OrchestratorValidationError(OrchestratorError):
    """Raised when local validation fails (bad args / bad time format)."""


class OrchestratorHTTPError(OrchestratorError):
    """Raised on non-2xx responses."""
    def __init__(self, status_code: int, message: str, body_preview: str = ""):
        super().__init__(f"{status_code}: {message}\n{body_preview}")
        self.status_code = status_code
        self.body_preview = body_preview


# ===========================
# HTTP helpers
# ===========================

def _build_session(retries: int, backoff: float) -> requests.Session:
    s = requests.Session()
    retry = Retry(
        total=retries,
        connect=retries,
        read=retries,
        status=retries,
        backoff_factor=backoff,
        status_forcelist=(429, 500, 502, 503, 504),
        allowed_methods=frozenset(["GET", "POST", "PUT", "DELETE"]),
        raise_on_status=False,
    )
    adapter = HTTPAdapter(max_retries=retry, pool_connections=10, pool_maxsize=20)
    s.mount("https://", adapter)
    s.mount("http://", adapter)
    return s


def _bearer_headers(token: str) -> Dict[str, str]:
    return {
        "Authorization": f"Bearer {token}",
        "Accept": "application/json",
    }


def _maybe_dict_list(objs: List[Any], return_fmt: Literal["dto", "dict"]) -> List[Any]:
    if return_fmt == "dict":
        return [o.to_dict() for o in objs]
    return objs


# ===========================
# Client (DTO or dict returns)
# ===========================

class OrchestratorClient:
    """
    Requests-based client with ergonomic returns:

      - Success -> typed DTOs (default) or plain dicts (return_fmt="dict")
      - Failure -> raises Orchestrator* exceptions

    Includes:
      - Dataclasses (SliceDTO, SliverDTO, PoaDataDTO) with .to_dict()
      - Structured logging (INFO per call; optional DEBUG dumps on failures)
    """

    TIME_FORMAT = "%Y-%m-%d %H:%M:%S %z"

    def __init__(
        self,
        orchestrator_host: str,
        *,
        timeout: float = 30.0,
        retries: int = 3,
        backoff_factor: float = 0.3,
        logger: Optional[logging.Logger] = None,
        http_debug: bool = False,
    ):
        if not orchestrator_host:
            raise OrchestratorValidationError("orchestrator_host must be specified")

        base_url = orchestrator_host
        if not base_url.startswith("http"):
            base_url = f"https://{base_url}"
        if not base_url.endswith("/"):
            base_url += "/"

        self.base_url = base_url
        self.timeout = timeout
        self.session = _build_session(retries=retries, backoff=backoff_factor)
        self.log = logger or logging.getLogger("fabric.orchestrator")
        self.http_debug = http_debug

    # ------------- Core HTTP -------------

    def _req(
        self,
        method: str,
        path: str,
        *,
        token: Optional[str] = None,
        params: Optional[Union[Dict[str, Any], List[Tuple[str, Any]]]] = None,
        json_body: Optional[Dict] = None,
        text_body: Optional[str] = None,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> requests.Response:
        url = self.base_url + path.lstrip("/")
        headers: Dict[str, str] = {}
        if token:
            headers.update(_bearer_headers(token))
        if extra_headers:
            headers.update(extra_headers)

        if text_body is not None and json_body is not None:
            raise OrchestratorValidationError("Provide either text_body or json_body, not both.")

        if json_body is not None and "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"
        if text_body is not None and "Content-Type" not in headers:
            headers["Content-Type"] = "text/plain"

        t0 = time.perf_counter()
        try:
            resp = self.session.request(
                method=method,
                url=url,
                headers=headers,
                params=params,
                json=json_body,
                data=text_body,
                timeout=self.timeout,
            )
            dt = (time.perf_counter() - t0) * 1000.0

            self.log.info(
                "orchestrator_http_call",
                extra={
                    "event": "http_call",
                    "http_method": method,
                    "url_path": path,
                    "status_code": resp.status_code,
                    "elapsed_ms": round(dt, 2),
                    "has_token": bool(token),
                    "params": params if isinstance(params, dict) else (params or []),
                    "payload_len": len(json.dumps(json_body)) if json_body is not None else (len(text_body) if text_body else 0),
                    "resp_len": len(resp.content or b""),
                },
            )

            if not (200 <= resp.status_code < 300):
                if self.http_debug:
                    self.log.debug(
                        "orchestrator_http_failure_dump",
                        extra={
                            "event": "http_failure",
                            "http_method": method,
                            "url": url,
                            "status_code": resp.status_code,
                            "request_headers": headers,
                            "request_params": params,
                            "request_json": json_body,
                            "request_text": text_body[:2000] if text_body else None,
                            "response_text": resp.text[:2000],
                        },
                    )
                raise OrchestratorHTTPError(resp.status_code, "HTTP request failed", resp.text[:2000])

            return resp
        except OrchestratorHTTPError:
            raise
        except Exception as e:
            dt = (time.perf_counter() - t0) * 1000.0
            self.log.error(
                "orchestrator_http_exception",
                extra={
                    "event": "http_exception",
                    "http_method": method,
                    "url_path": path,
                    "elapsed_ms": round(dt, 2),
                    "error": repr(e),
                },
            )
            raise OrchestratorError(repr(e)) from e

    @staticmethod
    def _json(resp: requests.Response) -> Dict[str, Any]:
        try:
            return resp.json() if resp.content else {}
        except Exception as e:
            raise OrchestratorError(f"Failed to parse JSON: {e}; status={resp.status_code}, text={resp.text[:500]}")

    # ------------- Public API (DTO or dict) -------------

    def create(
        self,
        *,
        token: str,
        slice_name: str,
        ssh_keys: Union[str, List[str]],
        topology: ExperimentTopology = None,
        slice_graph: str = None,
        lease_start_time: str = None,
        lease_end_time: str = None,
        lifetime: int = 24,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> List[SliverDTO] | List[Dict[str, Any]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        if not slice_name:
            raise OrchestratorValidationError("Slice Name must be specified")
        if (topology is None and slice_graph is None) or (topology is not None and slice_graph is not None):
            raise OrchestratorValidationError("Specify either topology or slice_graph")

        for lbl, ts in (("Lease Start Time", lease_start_time), ("Lease End Time", lease_end_time)):
            if ts:
                try:
                    datetime.strptime(ts, self.TIME_FORMAT)
                except Exception as e:
                    raise OrchestratorValidationError(f"{lbl} {ts} should be in format: {self.TIME_FORMAT} e: {e}")

        graph_str = slice_graph or topology.serialize()
        ssh_keys = [ssh_keys] if isinstance(ssh_keys, str) else list(ssh_keys)
        body = {"graph_model": graph_str, "ssh_keys": ssh_keys}

        params: Dict[str, Any] = {"name": slice_name, "lifetime": lifetime}
        if lease_start_time:
            params["lease_start_time"] = lease_start_time
        if lease_end_time:
            params["lease_end_time"] = lease_end_time

        resp = self._req("POST", "/slices/creates", token=token, params=params, json_body=body)
        data = self._json(resp).get("data") or []
        slivers = [SliverDTO.from_dict(x) for x in data]
        return _maybe_dict_list(slivers, return_fmt)

    def modify(
        self,
        *,
        token: str,
        slice_id: str,
        topology: ExperimentTopology = None,
        slice_graph: str = None,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> List[SliverDTO] | List[Dict[str, Any]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        if not slice_id:
            raise OrchestratorValidationError("Slice Id must be specified")
        if (topology is None and slice_graph is None) or (topology is not None and slice_graph is not None):
            raise OrchestratorValidationError("Specify either topology or slice_graph")

        graph_str = slice_graph or topology.serialize()
        resp = self._req(
            "PUT",
            f"/slices/modify/{slice_id}",
            token=token,
            text_body=graph_str,
            extra_headers={"Content-Type": "text/plain"},
        )
        data = self._json(resp).get("data") or []
        slivers = [SliverDTO.from_dict(x) for x in data]
        return _maybe_dict_list(slivers, return_fmt)

    def modify_accept(
        self,
        *,
        token: str,
        slice_id: str,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> Optional[ExperimentTopology] | Optional[Dict[str, str]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        if not slice_id:
            raise OrchestratorValidationError("Slice Id must be specified")

        resp = self._req("POST", f"/slices/modify/{slice_id}/accept", token=token)
        payload = self._json(resp)
        model = (payload.get("data") or [{}])[0].get("model")
        if not model:
            return None
        if return_fmt == "dict":
            return {"graph_format": GraphFormat.GRAPHML.name, "model": model}
        topo = ExperimentTopology()
        topo.load(graph_string=model)
        return topo

    def delete(self, *, token: str, slice_id: Optional[str] = None) -> None:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        path = f"/slices/delete/{slice_id}" if slice_id else "/slices/delete"
        self._req("DELETE", path, token=token)

    def slices(
        self,
        *,
        token: str,
        includes: Optional[List[str]] = None,  # state names (e.g., ["StableOK"])
        excludes: Optional[List[str]] = None,
        name: str = None,
        limit: int = 20,
        offset: int = 0,
        slice_id: str = None,
        as_self: bool = True,
        search: str = None,
        exact_match: bool = False,
        graph_format: str = GraphFormat.GRAPHML.name,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> List[SliceDTO] | List[Dict[str, Any]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")

        if slice_id:
            params = {"as_self": as_self, "graph_format": graph_format}
            resp = self._req("GET", f"/slices/{slice_id}", token=token, params=params)
            data = self._json(resp).get("data") or []
            slices = [SliceDTO.from_dict(x) for x in data]
            return _maybe_dict_list(slices, return_fmt)

        params: Dict[str, Union[str, int, bool]] = {
            "limit": limit,
            "offset": offset,
            "as_self": as_self,
            "exact_match": exact_match,
        }
        if name:
            params["name"] = name
        if search:
            params["search"] = search

        # API expects repeated "states" params if provided
        if includes or excludes:
            state_list = includes or [
                "StableError", "StableOK", "Nascent", "Configuring", "Closing", "Dead",
                "ModifyError", "ModifyOK", "Modifying", "AllocatedOK", "AllocatedError"
            ]
            if excludes:
                state_list = [s for s in state_list if s not in set(excludes)]

            params_list: List[Tuple[str, Any]] = list(params.items())
            for s in state_list:
                params_list.append(("states", s))
            resp = self._req("GET", "/slices", token=token, params=params_list)
        else:
            resp = self._req("GET", "/slices", token=token, params=params)

        data = self._json(resp).get("data") or []
        slices = [SliceDTO.from_dict(x) for x in data]
        return _maybe_dict_list(slices, return_fmt)

    def get_slice(
        self,
        *,
        token: str,
        slice_id: str,
        graph_format: str = str(GraphFormat.GRAPHML),
        as_self: bool = True,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> Optional[ExperimentTopology] | Optional[Dict[str, str]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        if not slice_id:
            raise OrchestratorValidationError("Slice Id must be specified")

        params = {"graph_format": graph_format, "as_self": as_self}
        resp = self._req("GET", f"/slices/{slice_id}", token=token, params=params)
        payload = self._json(resp)
        model = (payload.get("data") or [{}])[0].get("model")
        if not model:
            return None
        if return_fmt == "dict":
            return {"graph_format": graph_format, "model": model}
        topo = ExperimentTopology()
        topo.load(graph_string=model)
        return topo

    def slivers(
        self,
        *,
        token: str,
        slice_id: str,
        sliver_id: Optional[str] = None,
        as_self: bool = True,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> List[SliverDTO] | List[Dict[str, Any]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        if not slice_id:
            raise OrchestratorValidationError("Slice Id must be specified")

        params = {"slice_id": slice_id, "as_self": as_self}
        if sliver_id:
            resp = self._req("GET", f"/slivers/{sliver_id}", token=token, params=params)
        else:
            resp = self._req("GET", "/slivers", token=token, params=params)

        data = self._json(resp).get("data") or []
        slivers = [SliverDTO.from_dict(x) for x in data]
        return _maybe_dict_list(slivers, return_fmt)

    def resources(
        self,
        *,
        token: str,
        level: int = 1,
        force_refresh: bool = False,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        includes: Optional[List[str]] = None,
        excludes: Optional[List[str]] = None,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> Optional[AdvertizedTopology] | Optional[Dict[str, str]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")

        params: Dict[str, Union[str, int, bool]] = {
            "level": level,
            "force_refresh": force_refresh,
        }
        if start:
            params["start_date"] = start.strftime(self.TIME_FORMAT)
        if end:
            params["end_date"] = end.strftime(self.TIME_FORMAT)
        if includes:
            params["includes"] = ", ".join(includes)
        if excludes:
            params["excludes"] = ", ".join(excludes)

        resp = self._req("GET", "/resources", token=token, params=params)
        payload = self._json(resp)
        graph_string = (payload.get("data") or [{}])[0].get("model")
        if not graph_string:
            return None
        if return_fmt == "dict":
            return {"graph_format": GraphFormat.GRAPHML.name, "model": graph_string}
        substrate = AdvertizedTopology()
        substrate.load(graph_string=graph_string)
        return substrate

    def portal_resources(
        self,
        *,
        graph_format: str,
        level: Optional[int] = None,
        force_refresh: Optional[bool] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        includes: Optional[List[str]] = None,
        excludes: Optional[List[str]] = None,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> Optional[AdvertizedTopology] | Optional[Dict[str, str]]:
        params: Dict[str, Union[str, int, bool]] = {"graph_format": graph_format}
        if level is not None:
            params["level"] = level
        if force_refresh is not None:
            params["force_refresh"] = force_refresh
        if start:
            params["start_date"] = start.strftime(self.TIME_FORMAT)
        if end:
            params["end_date"] = end.strftime(self.TIME_FORMAT)
        if includes:
            params["includes"] = ", ".join(includes)
        if excludes:
            params["excludes"] = ", ".join(excludes)

        resp = self._req("GET", "/portalresources", params=params)
        payload = self._json(resp)
        graph_string = (payload.get("data") or [{}])[0].get("model")
        if not graph_string:
            return None
        if return_fmt == "dict":
            return {"graph_format": graph_format, "model": graph_string}
        substrate = AdvertizedTopology()
        substrate.load(graph_string=graph_string)
        return substrate

    def resources_summary(
        self,
        *,
        token: str,
        level: int = 2,
        force_refresh: bool = False,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        includes: Optional[List[str]] = None,
        excludes: Optional[List[str]] = None,
        resource_type: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Fetch a pre-built JSON resource summary from the orchestrator.
        Returns a dict with keys: sites, hosts, links, facility_ports.
        Much faster than resources() for clients that only need summary data.

        :param token: Fabric Identity Token
        :param level: Level of details (default 2)
        :param force_refresh: Force to retrieve current available resource information.
        :param start: start time
        :param end: end time
        :param includes: list of site names to include
        :param excludes: list of site names to exclude
        :param resource_type: comma-separated types to return (sites,hosts,links,facility_ports)
        :returns: dict with resource summary data
        """
        if not token:
            raise OrchestratorValidationError("Token must be specified")

        params: Dict[str, Union[str, int, bool]] = {
            "level": level,
            "force_refresh": force_refresh,
        }
        if start:
            params["start_date"] = start.strftime(self.TIME_FORMAT)
        if end:
            params["end_date"] = end.strftime(self.TIME_FORMAT)
        if includes:
            params["includes"] = ", ".join(includes)
        if excludes:
            params["excludes"] = ", ".join(excludes)
        if resource_type:
            params["type"] = resource_type

        resp = self._req("GET", "/resources/summary", token=token, params=params)
        payload = self._json(resp)
        data = (payload.get("data") or [{}])
        return data[0] if data else {}

    def portal_resources_summary(
        self,
        *,
        level: int = 2,
        force_refresh: bool = False,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        includes: Optional[List[str]] = None,
        excludes: Optional[List[str]] = None,
        resource_type: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Fetch a pre-built JSON resource summary from the orchestrator (portal, no auth).
        Returns a dict with keys: sites, hosts, links, facility_ports.

        :param level: Level of details (default 2)
        :param force_refresh: Force to retrieve current available resource information.
        :param start: start time
        :param end: end time
        :param includes: list of site names to include
        :param excludes: list of site names to exclude
        :param resource_type: comma-separated types to return (sites,hosts,links,facility_ports)
        :returns: dict with resource summary data
        """
        params: Dict[str, Union[str, int, bool]] = {
            "level": level,
            "force_refresh": force_refresh,
        }
        if start:
            params["start_date"] = start.strftime(self.TIME_FORMAT)
        if end:
            params["end_date"] = end.strftime(self.TIME_FORMAT)
        if includes:
            params["includes"] = ", ".join(includes)
        if excludes:
            params["excludes"] = ", ".join(excludes)
        if resource_type:
            params["type"] = resource_type

        resp = self._req("GET", "/portalresources/summary", params=params)
        payload = self._json(resp)
        data = (payload.get("data") or [{}])
        return data[0] if data else {}

    def renew(self, *, token: str, slice_id: str, new_lease_end_time: str) -> None:
        if not token or not slice_id or not new_lease_end_time:
            raise OrchestratorValidationError("Token, slice_id, and new_lease_end_time must be specified")
        try:
            datetime.strptime(new_lease_end_time, self.TIME_FORMAT)
        except Exception:
            raise OrchestratorValidationError(f"Lease End Time {new_lease_end_time} should be in format: {self.TIME_FORMAT}")

        params = {"lease_end_time": new_lease_end_time}
        self._req("POST", f"/slices/renew/{slice_id}", token=token, params=params)

    def poa(
        self,
        *,
        token: str,
        sliver_id: str,
        operation: str,
        vcpu_cpu_map: Optional[List[Dict[str, str]]] = None,
        node_set: Optional[List[str]] = None,
        keys: Optional[List[Dict[str, str]]] = None,
        bdf: Optional[List[str]] = None,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> List[PoaDataDTO] | List[Dict[str, Any]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        if not sliver_id:
            raise OrchestratorValidationError("Sliver Id must be specified")

        body: Dict[str, Union[str, Dict, List]] = {"operation": operation}
        pdata: Dict[str, Union[List, Dict]] = {}
        if vcpu_cpu_map is not None:
            pdata["vcpu_cpu_map"] = vcpu_cpu_map
        if node_set is not None:
            pdata["node_set"] = node_set
        if keys is not None:
            pdata["keys"] = keys
        if bdf is not None:
            pdata["bdf"] = bdf
        if pdata:
            body["data"] = pdata

        resp = self._req("POST", f"/poas/create/{sliver_id}", token=token, json_body=body)
        data = self._json(resp).get("data") or []
        poas = [PoaDataDTO.from_dict(x) for x in data]
        return _maybe_dict_list(poas, return_fmt)

    def get_poas(
        self,
        *,
        token: str,
        sliver_id: Optional[str] = None,
        poa_id: Optional[str] = None,
        limit: int = 20,
        offset: int = 0,
        return_fmt: Literal["dto", "dict"] = "dto",
    ) -> List[PoaDataDTO] | List[Dict[str, Any]]:
        if not token:
            raise OrchestratorValidationError("Token must be specified")
        if sliver_id is None and poa_id is None:
            raise OrchestratorValidationError("Sliver Id or Poa Id must be specified")

        if poa_id:
            resp = self._req("GET", f"/poas/{poa_id}", token=token)
        else:
            params: List[Tuple[str, Union[str, int]]] = [("limit", limit), ("offset", offset)]
            if sliver_id:
                params.append(("sliver_id", sliver_id))
            resp = self._req("GET", "/poas/", token=token, params=params)

        data = self._json(resp).get("data") or []
        poas = [PoaDataDTO.from_dict(x) for x in data]
        return _maybe_dict_list(poas, return_fmt)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Author Komal Thareja (kthare10@renci.org)

"""
High-level FABRIC manager that composes slice, user/project, and artifact operations.

This module defines :class:`FabricManager`, which extends your local
:class:`slice_manager.SliceManager` to add convenience methods for:

- User SSH key management (via Core API)
- User/project lookup (via Core API)
- Artifact CRUD and file upload/download (via Artifact Manager)

It honors **in-memory tokens** (``id_token`` / ``refresh_token``) for MCP-based
workflows and falls back to legacy file-based tokens when a token file path is
provided.
"""

from __future__ import annotations

import os
from typing import List, Optional, Tuple, Union

from fabric_cf.orchestrator.orchestrator_proxy import Status

from fabrictestbed.external_api.core_api import CoreApi
from fabrictestbed.external_api.artifact_manager import ArtifactManager, Visibility
from fabrictestbed.util.utils import Utils
from fabrictestbed.util.constants import Constants

from fabrictestbed.slice_manager import SliceManager


class FabricManagerException(Exception):
    """Custom exception for FabricManager-related errors."""
    pass


class FabricManager(SliceManager):
    """
    High-level orchestrator for FABRIC tasks across slices, users/projects, and artifacts.

    Inherits token handling and slice operations from :class:`slice_manager.SliceManager`,
    which itself uses an MCP-friendly :class:`token_manager.TokenManager`.

    :param cm_host: Credential Manager host (for token decoding / optional refresh).
    :type cm_host: str or None
    :param oc_host: Orchestrator host (e.g., ``orchestrator.fabric-testbed.net``).
                    If ``None``, read from ``FABRIC_ORCHESTRATOR_HOST`` env var.
    :type oc_host: str or None
    :param core_api_host: Core API host (e.g., ``alpha-6.fabric-testbed.net``).
                          If ``None``, read from ``FABRIC_CORE_API_HOST`` env var.
    :type core_api_host: str or None
    :param am_host: Artifact Manager base URL. If ``None``, read from ``FABRIC_AM_HOST`` env var.
    :type am_host: str or None
    :param token_location: Token JSON path (legacy mode). Use ``None`` for in-memory (MCP) mode.
    :type token_location: str or None
    :param project_id: Optional project UUID for scoping.
    :type project_id: str or None
    :param project_name: Optional project name for scoping.
    :type project_name: str or None
    :param scope: Token scope (default: ``"all"``) used if refresh is performed via CredMgr.
    :type scope: str
    :param id_token: In-memory FABRIC ID token (JWT) for MCP usage.
    :type id_token: str or None
    :param refresh_token: In-memory refresh token (optional) for MCP usage.
    :type refresh_token: str or None
    :param no_write: If ``True``, never write tokens to disk (recommended for MCP).
    :type no_write: bool
    :param auto_refresh: Flag to enable automatic token refresh
    :type auto_refresh: bool

    :raises FabricManagerException: If required env/hosts are not provided.
    """

    def __init__(
        self,
        *,
        cm_host: Optional[str] = None,
        oc_host: Optional[str] = None,
        core_api_host: Optional[str] = None,
        am_host: Optional[str] = None,
        token_location: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        scope: str = "all",
        id_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
        no_write: bool = False,
        auto_refresh: bool = True,
    ):
        super().__init__(
            cm_host=cm_host,
            oc_host=oc_host,
            token_location=token_location,
            project_id=project_id,
            project_name=project_name,
            scope=scope,
            id_token=id_token,
            refresh_token=refresh_token,
            no_write=no_write,
            auto_refresh=auto_refresh
        )

        if core_api_host is None:
            core_api_host = os.environ.get(Constants.FABRIC_CORE_API_HOST)
        if not core_api_host:
            raise FabricManagerException(f"Invalid initialization parameters: core_api_host={core_api_host!r}")
        self.core_api_host = core_api_host

        if am_host is None:
            am_host = os.environ.get(Constants.FABRIC_AM_HOST)
        if not am_host:
            raise FabricManagerException(f"Invalid initialization parameters: am_host={am_host!r}")
        self.am_host = am_host

    # -------------------------------------------------------------------------
    # SSH Keys (Core API)
    # -------------------------------------------------------------------------

    def get_ssh_keys(self, uuid: str = None, email: str = None) -> list:
        """
        Retrieve SSH keys associated with a user.

        :param uuid: User UUID to query (optional).
        :type uuid: str or None
        :param email: User email to query (optional).
        :type email: str or None
        :return: List of SSH key objects (shape as returned by Core API).
        :rtype: list
        :raises FabricManagerException: On Core API errors.
        """
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.get_ssh_keys(uuid=uuid, email=email)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def create_ssh_keys(
        self,
        *,
        key_type: str,
        description: str,
        comment: str = "ssh-key-via-api",
        store_pubkey: bool = True,
    ) -> list:
        """
        Create (register) SSH keys for the current user.

        :param key_type: Key type (e.g., ``"sliver"`` or ``"bastion"``).
        :type key_type: str
        :param description: Description to store for the key.
        :type description: str
        :param comment: Comment (often the public key's trailing comment).
        :type comment: str
        :param store_pubkey: If ``True``, store the public key.
        :type store_pubkey: bool
        :return: List of key records after creation (shape per Core API).
        :rtype: list
        :raises FabricManagerException: On Core API errors.
        """
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.create_ssh_keys(
                key_type=key_type,
                comment=comment,
                store_pubkey=store_pubkey,
                description=description,
            )
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    # -------------------------------------------------------------------------
    # Storage (Core API)
    # -------------------------------------------------------------------------

    def list_storage(self, offset: int = 0, limit: int = 200) -> list:
        """
        Retrieve all storage volumes.

        :param offset: Pagination offset (default: ``0``).
        :type offset: int
        :param limit: Maximum number of records to fetch (default: ``200``).
        :type limit: int
        :return: List of storage volume objects (shape as returned by Core API).
        :rtype: list
        :raises FabricManagerException: On Core API errors.
        """
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.list_storage(offset=offset, limit=limit)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def get_storage(self, uuid: str) -> list:
        """
        Retrieve a specific storage volume by UUID.

        :param uuid: Storage volume UUID.
        :type uuid: str
        :return: Storage volume details (shape as returned by Core API).
        :rtype: list
        :raises FabricManagerException: On Core API errors.
        """
        if not uuid:
            raise FabricManagerException("uuid must be provided")
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.get_storage(uuid=uuid)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    # -------------------------------------------------------------------------
    # User / Project Info (Core API)
    # -------------------------------------------------------------------------

    def get_user_info(self, *, uuid: str = None, email: str = None) -> dict:
        """
        Retrieve user info by email or uuid.

        :param uuid: User's uuid.
        :type uuid: str
        :param email: User's email address.
        :type email: str
        :return: User info dict (shape per Core API).
        :rtype: dict
        :raises FabricManagerException: On Core API errors.
        """
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.get_user_info(uuid=uuid, email=email)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def get_user_id(self) -> str:
        """
        Retrieve user id

        :return: User Guid
        :rtype: str
        :raises FabricManagerException: On Core API errors.
        """
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.get_user_id()
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def get_project_info(self, *, project_name: str = "all", project_id: str = "all", uuid: Optional[str] = None) -> list:
        """
        Retrieve project info for the current user.

        :param project_name: Project name filter (exact match in most deployments).
        :type project_name: str
        :param project_id: Project id filter (exact match in most deployments).
        :type project_id: str
        :param uuid: Optional user UUID; if omitted, the Core API may infer current user.
        :type uuid: str or None
        :return: List of matching project records (shape per Core API).
        :rtype: list
        :raises FabricManagerException: On Core API errors.
        """
        if not project_name and not project_id:
            raise FabricManagerException("project_name or project_id must be provided")
        try:
            core_api_proxy = CoreApi(core_api_host=self.core_api_host, token=self.ensure_valid_id_token())
            return core_api_proxy.get_user_projects(uuid=uuid, project_name=project_name, project_id=project_id)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def get_metrics_overview(self,
                             excluded_projects: List[str] = None,
                             authenticated: bool = False
                             ) -> Tuple[Status, Union[list, Exception]]:
        """
        Fetch aggregate Core API metrics/overview.

        :param excluded_projects: List of projects to exclude from the metrics
        :param authenticated: Flag indicating whether to retrieve metrics for a specific user
        :return: Tuple containing the status and a list of metrics or an exception
        """
        try:
            token = self.ensure_valid_id_token() if authenticated else None
            return self.oc_proxy.get_metrics_overview(token=token,
                                                      excluded_projects=excluded_projects)

        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def create_artifact(self,
                        artifact_title: str,
                        description_short: str,
                        description_long: str,
                        authors: List[str],
                        tags: List[str],
                        visibility: Visibility = Visibility.Author,
                        update_existing: bool = True
                        ) -> dict:
        """
        Create a new artifact.

        :param artifact_title: Title of the artifact
        :param description_short: Short description of the artifact
        :param description_long: Long description of the artifact
        :param authors: List of authors associated with the artifact
        :param tags: List of tags associated with the artifact
        :param visibility: Visibility level of the artifact
        :param update_existing: Flag indicating whether to update an existing artifact
        :return: Dictionary containing the artifact details
        :raises FabricManagerException: If there is an error in creating or updating the artifact.
        """
        try:
            if not authors:
                authors = []
            am_proxy = ArtifactManager(api_url=self.am_host, token=self.ensure_valid_id_token())
            existing_artifacts = am_proxy.list_artifacts(search=artifact_title)

            artifact = None
            if update_existing:
                for e in existing_artifacts:
                    for author in e.get("authors"):
                        if author.get("uuid") == self.get_user_id():
                            artifact = e
                            break
                    if artifact:
                        break

            author_ids = [self.get_user_id()]
            if self.user_email in authors:
                authors.remove(self.user_email)
            for a in authors:
                author_info = self.get_user_info(email=a)
                if author_info and author_info.get('uuid') and author_info.get('uuid') not in author_ids:
                    author_ids.append(author_info.get('uuid'))

            if not artifact:
                artifact = am_proxy.create_artifact(
                    artifact_title=artifact_title,
                    description_short=description_short,
                    description_long=description_long,
                    tags=tags,
                    visibility=visibility,
                    authors=author_ids,
                    project_id=self.project_id
                )
            else:
                am_proxy.update_artifact(artifact_id=artifact.get("uuid"),
                                         artifact_title=artifact_title,
                                         description_short=description_short,
                                         description_long=description_long,
                                         tags=tags,
                                         visibility=visibility,
                                         authors=author_ids, project_id=self.project_id)
            return artifact

        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def list_artifacts(self,
                       search: str = None,
                       artifact_id: str = None) -> list:
        """
        List artifacts with optional filters.

        :param artifact_id: Specific artifact UUID filter.
        :type artifact_id: str or None
        :param search: Search string (title substring match).
        :type search: str or None
        :return: List of artifact records.
        :rtype: list
        :raises FabricManagerException: On Artifact Manager errors.
        """
        try:
            am_proxy = ArtifactManager(api_url=self.am_host, token=self.ensure_valid_id_token())
            if not artifact_id:
                return am_proxy.list_artifacts(search=search)
            else:
                return [am_proxy.get_artifact(artifact_id=artifact_id)]
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def delete_artifact(self, *, artifact_id: str) -> str:
        """
        Delete an artifact by UUID.

        :param artifact_id: Artifact UUID to delete.
        :type artifact_id: str
        :return: Confirmation message from Artifact Manager.
        :rtype: str
        :raises FabricManagerException: On Artifact Manager errors.
        """
        if not artifact_id:
            raise FabricManagerException("artifact_id must be provided")
        try:
            am_proxy = ArtifactManager(api_url=self.am_host, token=self.ensure_valid_id_token())
            return am_proxy.delete_artifact(artifact_id=artifact_id)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def get_tags(self) -> list:
        """
        Retrieve available artifact tags.

        :return: List of tag strings.
        :rtype: list
        :raises FabricManagerException: On Artifact Manager errors.
        """
        try:
            am_proxy = ArtifactManager(api_url=self.am_host, token=self.ensure_valid_id_token())
            return am_proxy.get_tags()
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def upload_file_to_artifact(self,
                                file_to_upload: str,
                                artifact_id: str = None,
                                artifact_title: str = None) -> dict:
        """
        Upload a file as a new artifact version.

        This method uploads a file to an artifact identified by either its `artifact_id` or `artifact_title`.
        If `artifact_id` is not provided, the method will search for the artifact using `artifact_title` before uploading the file.

        :param file_to_upload: The path to the file that should be uploaded.
        :param artifact_id: The unique identifier of the artifact to which the file will be uploaded.
        :param artifact_title: The title of the artifact to which the file will be uploaded.
        :return: A dictionary containing the details of the uploaded file.
        :raises ValueError: If neither `artifact_id` nor `artifact_title` is provided.
        :raises FabricManagerException: If an error occurs during the upload process.
        """
        if artifact_id is None and artifact_title is None:
            raise ValueError("Either artifact_id or artifact_title must be specified!")

        try:
            am_proxy = ArtifactManager(api_url=self.am_host, token=self.ensure_valid_id_token())
            artifacts = self.list_artifacts(artifact_id=artifact_id, search=artifact_title)
            if len(artifacts) != 1:
                raise ValueError(f"Requested artifact: {artifact_id}/{artifact_title} has 0 or more than versions "
                                 f"available, Please specify the version to download!")
            artifact = artifacts[0]
            artifact_id = artifact.get("uuid")
            return am_proxy.upload_file_to_artifact(artifact_id=artifact_id, file_path=file_to_upload)
        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)

    def download_artifact(
        self,
        *,
        download_dir: str,
        artifact_id: Optional[str] = None,
        artifact_title: Optional[str] = None,
        version: Optional[str] = None,
        version_urn: Optional[str] = None,
    ) -> str:
        """
        Download an artifact (by id/title/version or a specific version URN).

        Provide either ``artifact_id`` or ``artifact_title`` (with optional ``version``),
        **or** pass a concrete ``version_urn`` for direct download.

        :param download_dir: Destination directory for the downloaded file(s).
        :type download_dir: str
        :param artifact_id: Artifact UUID to resolve (optional).
        :type artifact_id: str or None
        :param artifact_title: Artifact title to search (optional).
        :type artifact_title: str or None
        :param version: Specific version label to download (optional).
        :type version: str or None
        :param version_urn: Exact version URN to download (bypasses listing).
        :type version_urn: str or None
        :return: Path to the downloaded file (or directory), per Artifact Manager behavior.
        :rtype: str
        :raises ValueError: If neither an artifact identifier nor a version URN is provided.
        :raises FabricManagerException: On Artifact Manager errors.
        """
        if not download_dir:
            raise FabricManagerException("download_dir must be provided")

        if artifact_id is None and artifact_title is None and version_urn is None:
            raise ValueError("Either artifact_id, artifact_title, or version_urn must be specified!")

        try:
            am_proxy = ArtifactManager(api_url=self.am_host, token=self.ensure_valid_id_token())

            if not version_urn:
                # Resolve to a specific version URN
                artifacts = self.list_artifacts(artifact_id=artifact_id, search=artifact_title)
                if len(artifacts) != 1:
                    raise ValueError(
                        f"Ambiguous artifact resolution for id={artifact_id!r} title={artifact_title!r}: "
                        f"expected exactly one artifact, got {len(artifacts)}"
                    )
                artifact = artifacts[0]
                versions = artifact.get("versions", []) or []
                version_urn = None

                if version:
                    # Find matching version entry by label
                    for v in versions:
                        if v.get("version") == version:
                            version_urn = v.get("urn")
                            break
                    if not version_urn:
                        raise ValueError(f"Version {version!r} not found for artifact {artifact.get('uuid')}")
                else:
                    # Choose the first/primary version if no version specified
                    if versions:
                        version_urn = versions[0].get("urn")

                if not version_urn:
                    raise ValueError("Could not resolve a version URN for the requested artifact")

            # Perform the download by URN
            return am_proxy.download_artifact(urn=version_urn, download_dir=download_dir, version=version)

        except Exception as e:
            error_message = Utils.extract_error_message(exception=e)
            raise FabricManagerException(error_message)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
#
import hashlib
import json
from datetime import datetime, timedelta

from fss_utils.jwt_manager import ValidateCode
from fss_utils.jwt_validate import JWTValidator


class Utils:
    @staticmethod
    def generate_sha256(*, token: str):
        """
        Generate SHA 256 for a token
        @param token token string
        """
        # Create a new SHA256 hash object
        sha256_hash = hashlib.sha256()

        # Convert the string to bytes and update the hash object
        sha256_hash.update(token.encode('utf-8'))

        # Get the hexadecimal representation of the hash
        sha256_hex = sha256_hash.hexdigest()

        return sha256_hex

    @staticmethod
    def extract_error_message(*, exception):
        if hasattr(exception, "body"):
            response_body = exception.body
        elif hasattr(exception, "error"):
            response_body = exception.error
        elif isinstance(exception, dict) and "error" in exception:
            response_body = exception.get("error")
        else:
            response_body = exception
        if response_body:
            try:
                if not isinstance(response_body, dict):
                    response_body = json.loads(response_body)
                errors = response_body.get("errors")
                if errors and len(errors) > 0:
                    return f"{errors[0].get('message')} - {errors[0].get('details')}"
            except Exception:
                return str(exception)
        return str(exception)

    @staticmethod
    def decode_token(*, cm_host: str, token: str) -> dict:
        t = datetime.strptime("00:10:00", "%H:%M:%S")
        jwt_validator = JWTValidator(url=f"https://{cm_host}/credmgr/certs",
                                     refresh_period=timedelta(hours=t.hour, minutes=t.minute, seconds=t.second))
        code, token_or_exception = jwt_validator.validate_jwt(token=token, verify_exp=True)
        if code is not ValidateCode.VALID:
            raise Exception(f"Unable to validate provided token: {code}/{token_or_exception}")
        return token_or_exception
